# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Google Cloud Platform library - %%arguments IPython Cell Magic Functionality."""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
from past.builtins import basestring

try:
  import IPython
  import IPython.core.magic
except ImportError:
  raise Exception('This module can only be loaded in ipython.')

import argparse
import datetime
import imp
import re
import sys
import time

import datalab.bigquery
import datalab.data
from datalab.utils.commands import CommandParser, handle_magic_line


def _create_sql_parser():
  sql_parser = CommandParser(prog="%%sql",
                             formatter_class=argparse.RawDescriptionHelpFormatter,
                             description="""
Create a named SQL module with one or more queries.

The cell body should contain an optional initial part defining the default
values for the variables, if any, using Python code, followed by one or more
queries.

Queries should start with 'DEFINE QUERY <name>' in order to bind them to
<module name>.<query name> in the notebook (as datalab.data.SqlStament instances).
The final query can optionally omit 'DEFINE QUERY <name>', as using the module
name in places where a SqlStatement is expected will resolve to the final query
in the module.

Queries can refer to variables with '$<name>', as well as refer to other queries
within the same module, making it easy to compose nested queries and test their
parts.

The Python code defining the variable default values can assign scalar or list/tuple values to
variables, or one of the special functions 'datestring' and 'source'.

When a variable with a 'datestring' default is expanded it will expand to a formatted
string based on the current date, while a 'source' default will expand to a table whose
name is based on the current date.

datestring() takes two named arguments, 'format' and 'offset'. The former is a
format string that is the same as for Python's time.strftime function. The latter
is a string containing a comma-separated list of expressions such as -1y, +2m,
etc; these are offsets from the time of expansion that are applied in order. The
suffix (y, m, d, h, M) correspond to units of years, months, days, hours and
minutes, while the +n or -n prefix is the number of units to add or subtract from
the time of expansion. Three special values 'now', 'today' and 'yesterday' are
also supported; 'today' and 'yesterday' will be midnight UTC on the current date
or previous days date.

source() can take a 'name' argument for a fixed table name, or 'format' and 'offset'
arguments similar to datestring(), but unlike datestring() will resolve to a Table
with the specified name.
""")
  sql_parser.add_argument('-m', '--module', help='The name for this SQL module')
  sql_parser.add_argument('-d', '--dialect', help='BigQuery SQL dialect',
                             choices=['legacy', 'standard'])
  sql_parser.add_argument('-b', '--billing', type=int, help='BigQuery billing tier')
  sql_parser.set_defaults(func=lambda args, cell: sql_cell(args, cell))
  return sql_parser


_sql_parser = _create_sql_parser()


# Register the line magic as well as the cell magic so we can at least give people help
# without requiring them to enter cell content first.
@IPython.core.magic.register_line_cell_magic
def sql(line, cell=None):
  """ Create a SQL module with one or more queries. Use %sql --help for more details.

  The supported syntax is:

  %%sql [--module <modulename>]
  [<optional Python code for default argument values>]
  [<optional named queries>]
  [<optional unnamed query>]

  At least one query should be present. Named queries should start with:

    DEFINE QUERY <name>

  on a line by itself.

  Args:
  args: the optional arguments following '%%sql'.
  cell: the contents of the cell; Python code for arguments followed by SQL queries.
   """
  if cell is None:
    _sql_parser.print_help()
  else:
    return handle_magic_line(line, cell, _sql_parser)


def _date(val, offset=None):
  """ A special pseudo-type for pipeline arguments.

  This allows us to parse dates as Python datetimes, including special values like 'now'
  and 'today', as well as apply offsets to the datetime.

  Args:
    val: a string containing the value for the datetime. This can be 'now', 'today' (midnight at
        start of day), 'yesterday' (midnight at start of yesterday), or a formatted date that
        will be passed to the datetime constructor. Note that 'now' etc are assumed to
        be in UTC.
    offset: for date arguments a string containing a comma-separated list of
      relative offsets to apply of the form <n><u> where <n> is an integer and
      <u> is a single character unit (d=day, m=month, y=year, h=hour, m=minute).

  Returns:
    A Python datetime resulting from starting at <val> and applying the sequence of deltas
    specified in <offset>.
  """
  if val is None:
    return val
  if val == '' or val == 'now':
    when = datetime.datetime.utcnow()
  elif val == 'today':
    dt = datetime.datetime.utcnow()
    when = datetime.datetime(dt.year, dt.month, dt.day)
  elif val == 'yesterday':
    dt = datetime.datetime.utcnow() - datetime.timedelta(1)
    when = datetime.datetime(dt.year, dt.month, dt.day)
  else:
    when = datetime.datetime.strptime(val, "%Y%m%d")
  if offset is not None:
    for part in offset.split(','):
      unit = part[-1]
      quantity = int(part[:-1])
      # We can use timedelta for days and under, but not for years and months
      if unit == 'y':
        when = datetime.datetime(year=when.year + quantity, month=when.month, day=when.day,
                                 hour=when.hour, minute=when.minute)
      elif unit == 'm':
        new_year = when.year
        new_month = when.month + quantity
        if new_month < 1:
          new_month = -new_month
          new_year += 1 + (new_month // 12)
          new_month = 12 - new_month % 12
        elif new_month > 12:
          new_year += (new_month - 1) // 12
          new_month = 1 + (new_month - 1) % 12
        when = datetime.datetime(year=new_year, month=new_month, day=when.day,
                                 hour=when.hour, minute=when.minute)
      elif unit == 'd':
        when += datetime.timedelta(days=quantity)
      elif unit == 'h':
        when += datetime.timedelta(hours=quantity)
      elif unit == 'M':
        when += datetime.timedelta(minutes=quantity)

  return when


def _resolve_table(v, format, delta):
  try:
    when = _date(v, delta)
    v = time.strftime(format, when.timetuple())
  except Exception:
    pass
  return datalab.bigquery.Table(v)


def _make_string_formatter(f, offset=None):
  """ A closure-izer for string arguments that include a format and possibly an offset. """
  format = f
  delta = offset
  return lambda v: time.strftime(format, (_date(v, delta)).timetuple())


def _make_table_formatter(f, offset=None):
  """ A closure-izer for table arguments that include a format and possibly an offset. """
  format = f
  delta = offset
  return lambda v: _resolve_table(v, format, delta)


def _make_table(v):
  return datalab.bigquery.Table(v)


def _datestring(format, offset=''):
  return {'type': 'datestring', 'format': format, 'offset': offset}


def _table(name=None, format=None, offset=''):
  return {'type': 'table', 'name': name, 'format': format, 'offset': offset}


def _arguments(code, module):
  """Define pipeline arguments.

  Args:
    code: the Python code to execute that defines the arguments.

  """
  arg_parser = CommandParser.create('')
  try:
    # Define our special argument 'types' and add them to the environment.
    builtins = {'source': _table, 'datestring': _datestring}
    env = {}
    env.update(builtins)

    # Execute the cell which should be one or more calls to arg().
    exec(code, env)

    # Iterate through the module dictionary. For any newly defined objects,
    # add args to the parser.
    for key in env:

      # Skip internal/private stuff.
      if key in builtins or key[0] == '_':
        continue
      # If we want to support importing query modules into other query modules, uncomment next 4
      # Skip imports but add them to the module
      # if isinstance(env[key], types.ModuleType):
      #   module.__dict__[key] = env[key]
      #   continue

      val = env[key]
      key = '--%s' % key

      if isinstance(val, bool):
        if val:
          arg_parser.add_argument(key, default=val, action='store_true')
        else:
          arg_parser.add_argument(key, default=val, action='store_false')
      elif isinstance(val, basestring) or isinstance(val, int) or isinstance(val, float) \
          or isinstance(val, int):
        arg_parser.add_argument(key, default=val)
      elif isinstance(val, list):
        arg_parser.add_argument(key, default=val, nargs='+')
      elif isinstance(val, tuple):
        arg_parser.add_argument(key, default=list(val), nargs='+')

      # Is this one of our pseudo-types for dates/tables?
      elif isinstance(val, dict) and 'type' in val:
        if val['type'] == 'datestring':
          arg_parser.add_argument(key, default='',
                                  type=_make_string_formatter(val['format'],
                                                              offset=val['offset']))
        elif val['type'] == 'table':
          if val['format'] is not None:
            arg_parser.add_argument(key, default='',
                                    type=_make_table_formatter(val['format'],
                                                               offset=val['offset']))
          else:
            arg_parser.add_argument(key, default=val['name'], type=_make_table)
        else:
          raise Exception('Cannot generate argument for %s of type %s' % (key, type(val)))
      else:
        raise Exception('Cannot generate argument for %s of type %s' % (key, type(val)))

  except Exception as e:
    print("%%sql arguments: %s from code '%s'" % (str(e), str(code)))
  return arg_parser


def _split_cell(cell, module):
  """ Split a hybrid %%sql cell into the Python code and the queries.

    Populates a module with the queries.

  Args:
    cell: the contents of the %%sql cell.
    module: the module that the contents will populate.

  Returns:
    The default (last) query for the module.

  """
  lines = cell.split('\n')
  code = None
  last_def = -1
  name = None
  define_wild_re = re.compile('^DEFINE\s+.*$', re.IGNORECASE)
  define_re = re.compile('^DEFINE\s+QUERY\s+([A-Z]\w*)\s*?(.*)$', re.IGNORECASE)
  select_re = re.compile('^SELECT\s*.*$', re.IGNORECASE)
  standard_sql_re = re.compile('^(CREATE|WITH|INSERT|DELETE|UPDATE)\s*.*$', re.IGNORECASE)
  # TODO(gram): a potential issue with this code is if we have leading Python code followed
  # by a SQL-style comment before we see SELECT/DEFINE. When switching to the tokenizer see
  # if we can address this.
  for i, line in enumerate(lines):
    define_match = define_re.match(line)
    select_match = select_re.match(line)
    standard_sql_match = standard_sql_re.match(line)

    if i:
      prior_content = ''.join(lines[:i]).strip()
      if select_match:
        # Avoid matching if previous token was '(' or if Standard SQL is found
        # TODO: handle the possibility of comments immediately preceding SELECT
        select_match = len(prior_content) == 0 or (prior_content[-1] != '(' and not standard_sql_re.match(prior_content))
      if standard_sql_match:
        standard_sql_match = len(prior_content) == 0 or not standard_sql_re.match(prior_content)

    if define_match or select_match or standard_sql_match:
      # If this is the first query, get the preceding Python code.
      if code is None:
        code = ('\n'.join(lines[:i])).strip()
        if len(code):
          code += '\n'
      elif last_def >= 0:

        # This is not the first query, so gather the previous query text.
        query = '\n'.join([line for line in lines[last_def:i] if len(line)]).strip()
        if select_match and name != datalab.data._utils._SQL_MODULE_MAIN and len(query) == 0:
          # Avoid DEFINE query name\nSELECT ... being seen as an empty DEFINE followed by SELECT
          continue

        # Save the query
        statement = datalab.data.SqlStatement(query, module)
        module.__dict__[name] = statement
        # And set the 'last' query to be this too
        module.__dict__[datalab.data._utils._SQL_MODULE_LAST] = statement

      # Get the query name and strip off our syntactic sugar if appropriate.
      if define_match:
        name = define_match.group(1)
        lines[i] = define_match.group(2)
      else:
        name = datalab.data._utils._SQL_MODULE_MAIN

      # Save the starting line index of the new query
      last_def = i
    else:
      define_wild_match = define_wild_re.match(line)
      if define_wild_match:
        raise Exception('Expected "DEFINE QUERY <name>"')

  if last_def >= 0:
    # We were in a query so save this tail query.
    query = '\n'.join([line for line in lines[last_def:] if len(line)]).strip()
    statement = datalab.data.SqlStatement(query, module)
    module.__dict__[name] = statement
    module.__dict__[datalab.data._utils._SQL_MODULE_LAST] = statement

  if code is None:
    code = ''
  module.__dict__[datalab.data._utils._SQL_MODULE_ARGPARSE] = _arguments(code, module)
  return module.__dict__.get(datalab.data._utils._SQL_MODULE_LAST, None)


def sql_cell(args, cell):
  """Implements the SQL cell magic for ipython notebooks.

  The supported syntax is:

      %%sql [--module <modulename>]
      [<optional Python code for default argument values>]
      [<optional named queries>]
      [<optional unnamed query>]

  At least one query should be present. Named queries should start with:

      DEFINE QUERY <name>

  on a line by itself.

  Args:
    args: the optional arguments following '%%sql'.
    cell: the contents of the cell; Python code for arguments followed by SQL queries.
  """
  name = args['module'] if args['module'] else '_sql_cell'
  module = imp.new_module(name)
  query = _split_cell(cell, module)
  ipy = IPython.get_ipython()
  if not args['module']:
      # Execute now
      if query:
        return datalab.bigquery.Query(query, values=ipy.user_ns) \
          .execute(dialect=args['dialect'], billing_tier=args['billing']).results
  else:
    # Add it as a module
    sys.modules[name] = module
    exec('import %s' % name, ipy.user_ns)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements CSV options for External Tables and Table loads from GCS."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import object


class CSVOptions(object):

  def __init__(self, delimiter=',', skip_leading_rows=0, encoding='utf-8', quote='"',
               allow_quoted_newlines=False, allow_jagged_rows=False):

    """ Initialize an instance of CSV options.

    Args:
      delimiter: The separator for fields in a CSV file. BigQuery converts the string to
          ISO-8859-1 encoding, and then uses the first byte of the encoded string to split the data
          as raw binary (default ',').
      skip_leading_rows: A number of rows at the top of a CSV file to skip (default 0).
      encoding: The character encoding of the data, either 'utf-8' (the default) or 'iso-8859-1'.
      quote: The value used to quote data sections in a CSV file; default '"'. If your data does
          not contain quoted sections, set the property value to an empty string. If your data
          contains quoted newline characters, you must also enable allow_quoted_newlines.
      allow_quoted_newlines: If True, allow quoted data sections in CSV files that contain newline
          characters (default False).
      allow_jagged_rows: If True, accept rows in CSV files that are missing trailing optional
          columns; the missing values are treated as nulls (default False).
    """
    encoding_upper = encoding.upper()
    if encoding_upper != 'UTF-8' and encoding_upper != 'ISO-8859-1':
      raise Exception("Invalid source encoding %s" % encoding)

    self._delimiter = delimiter
    self._skip_leading_rows = skip_leading_rows
    self._encoding = encoding
    self._quote = quote
    self._allow_quoted_newlines = allow_quoted_newlines
    self._allow_jagged_rows = allow_jagged_rows

  @property
  def delimiter(self):
    return self._delimiter

  @property
  def skip_leading_rows(self):
    return self._skip_leading_rows

  @property
  def encoding(self):
    return self._encoding

  @property
  def quote(self):
      return self._quote

  @property
  def allow_quoted_newlines(self):
    return self._allow_quoted_newlines

  @property
  def allow_jagged_rows(self):
    return self._allow_jagged_rows

  def _to_query_json(self):
    """ Return the options as a dictionary to be used as JSON in a query job. """
    return {
      'quote': self._quote,
      'fieldDelimiter': self._delimiter,
      'encoding': self._encoding.upper(),
      'skipLeadingRows': self._skip_leading_rows,
      'allowQuotedNewlines': self._allow_quoted_newlines,
      'allowJaggedRows': self._allow_jagged_rows
    }



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Google Cloud Platform library - BigQuery IPython Functionality."""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from builtins import zip
from builtins import str
from past.builtins import basestring

try:
  import IPython
  import IPython.core.display
  import IPython.core.magic
except ImportError:
  raise Exception('This module can only be loaded in ipython.')

import fnmatch
import json
import re

import datalab.bigquery
import datalab.data
import datalab.utils
import datalab.utils.commands


def _create_create_subparser(parser):
  create_parser = parser.subcommand('create', 'Create a dataset or table.')
  sub_commands = create_parser.add_subparsers(dest='command')

  create_dataset_parser = sub_commands.add_parser('dataset', help='Create a dataset.')
  create_dataset_parser.add_argument('-n', '--name', help='The name of the dataset to create.',
                                     required=True)
  create_dataset_parser.add_argument('-f', '--friendly', help='The friendly name of the dataset.')

  create_table_parser = sub_commands.add_parser('table', help='Create a table.')
  create_table_parser.add_argument('-n', '--name', help='The name of the table to create.',
                                   required=True)
  create_table_parser.add_argument('-o', '--overwrite', help='Overwrite table if it exists.',
                                   action='store_true')
  return create_parser


def _create_delete_subparser(parser):
  delete_parser = parser.subcommand('delete', 'Delete a dataset or table.')
  sub_commands = delete_parser.add_subparsers(dest='command')

  delete_dataset_parser = sub_commands.add_parser('dataset', help='Delete a dataset.')
  delete_dataset_parser.add_argument('-n', '--name', help='The name of the dataset to delete.',
                                     required=True)

  delete_table_parser = sub_commands.add_parser('table', help='Delete a table.')
  delete_table_parser.add_argument('-n', '--name', help='The name of the table to delete.',
                                   required=True)
  return delete_parser


def _create_sample_subparser(parser):
  sample_parser = parser.subcommand('sample',
      'Display a sample of the results of a BigQuery SQL query.\n' +
      'The cell can optionally contain arguments for expanding variables in the query,\n' +
      'if -q/--query was used, or it can contain SQL for a query.')
  group = sample_parser.add_mutually_exclusive_group()
  group.add_argument('-q', '--query', help='the name of the query to sample')
  group.add_argument('-t', '--table', help='the name of the table to sample')
  group.add_argument('-v', '--view', help='the name of the view to sample')
  sample_parser.add_argument('-d', '--dialect', help='BigQuery SQL dialect',
                             choices=['legacy', 'standard'])
  sample_parser.add_argument('-b', '--billing', type=int, help='BigQuery billing tier')
  sample_parser.add_argument('-c', '--count', type=int, default=10,
                             help='The number of rows to limit to, if sampling')
  sample_parser.add_argument('-m', '--method', help='The type of sampling to use',
                             choices=['limit', 'random', 'hashed', 'sorted'], default='limit')
  sample_parser.add_argument('-p', '--percent', type=int, default=1,
                             help='For random or hashed sampling, what percentage to sample from')
  sample_parser.add_argument('-f', '--field',
                             help='The field to use for sorted or hashed sampling')
  sample_parser.add_argument('-o', '--order', choices=['ascending', 'descending'],
                             default='ascending', help='The sort order to use for sorted sampling')
  sample_parser.add_argument('-P', '--profile', action='store_true',
                             default=False, help='Generate an interactive profile of the data')
  sample_parser.add_argument('--verbose',
                             help='Show the expanded SQL that is being executed',
                             action='store_true')
  return sample_parser


def _create_udf_subparser(parser):
  udf_parser = parser.subcommand('udf', 'Create a named Javascript BigQuery UDF')
  udf_parser.add_argument('-m', '--module', help='The name for this UDF')
  return udf_parser


def _create_dry_run_subparser(parser):
  dry_run_parser = parser.subcommand('dryrun',
      'Execute a dry run of a BigQuery query and display approximate usage statistics')
  dry_run_parser.add_argument('-q', '--query',
                              help='The name of the query to be dry run')
  dry_run_parser.add_argument('-d', '--dialect', help='BigQuery SQL dialect',
                             choices=['legacy', 'standard'])
  dry_run_parser.add_argument('-b', '--billing', type=int, help='BigQuery billing tier')
  dry_run_parser.add_argument('-v', '--verbose',
                              help='Show the expanded SQL that is being executed',
                              action='store_true')
  return dry_run_parser


def _create_execute_subparser(parser):
  execute_parser = parser.subcommand('execute',
      'Execute a BigQuery SQL query and optionally send the results to a named table.\n' +
      'The cell can optionally contain arguments for expanding variables in the query.')
  execute_parser.add_argument('-nc', '--nocache', help='Don\'t use previously cached results',
                              action='store_true')
  execute_parser.add_argument('-d', '--dialect', help='BigQuery SQL dialect',
                             choices=['legacy', 'standard'])
  execute_parser.add_argument('-b', '--billing', type=int, help='BigQuery billing tier')
  execute_parser.add_argument('-m', '--mode', help='The table creation mode', default='create',
                              choices=['create', 'append', 'overwrite'])
  execute_parser.add_argument('-l', '--large', help='Whether to allow large results',
                              action='store_true')
  execute_parser.add_argument('-q', '--query', help='The name of query to run')
  execute_parser.add_argument('-t', '--target', help='target table name')
  execute_parser.add_argument('-v', '--verbose',
                              help='Show the expanded SQL that is being executed',
                              action='store_true')
  return execute_parser


def _create_pipeline_subparser(parser):
  pipeline_parser = parser.subcommand('pipeline',
      'Define a deployable pipeline based on a BigQuery query.\n' +
      'The cell can optionally contain arguments for expanding variables in the query.')
  pipeline_parser.add_argument('-n', '--name', help='The pipeline name')
  pipeline_parser.add_argument('-nc', '--nocache', help='Don\'t use previously cached results',
                               action='store_true')
  pipeline_parser.add_argument('-d', '--dialect', help='BigQuery SQL dialect',
                             choices=['legacy', 'standard'])
  pipeline_parser.add_argument('-b', '--billing', type=int, help='BigQuery billing tier')
  pipeline_parser.add_argument('-m', '--mode', help='The table creation mode', default='create',
                               choices=['create', 'append', 'overwrite'])
  pipeline_parser.add_argument('-l', '--large', help='Allow large results', action='store_true')
  pipeline_parser.add_argument('-q', '--query', help='The name of query to run', required=True)
  pipeline_parser.add_argument('-t', '--target', help='The target table name', nargs='?')
  pipeline_parser.add_argument('-v', '--verbose',
                               help='Show the expanded SQL that is being executed',
                               action='store_true')
  pipeline_parser.add_argument('action', nargs='?', choices=('deploy', 'run', 'dryrun'),
                               default='dryrun',
                               help='Whether to deploy the pipeline, execute it immediately in ' +
                                    'the notebook, or validate it with a dry run')
  # TODO(gram): we may want to move some command line arguments to the cell body config spec
  # eventually.
  return pipeline_parser


def _create_table_subparser(parser):
  table_parser = parser.subcommand('table', 'View a BigQuery table.')
  table_parser.add_argument('-r', '--rows', type=int, default=25,
                            help='Rows to display per page')
  table_parser.add_argument('-c', '--cols',
                            help='Comma-separated list of column names to restrict to')
  table_parser.add_argument('table', help='The name of, or a reference to, the table or view')
  return table_parser


def _create_schema_subparser(parser):
  schema_parser = parser.subcommand('schema', 'View a BigQuery table or view schema.')
  group = schema_parser.add_mutually_exclusive_group()
  group.add_argument('-v', '--view', help='the name of the view whose schema should be displayed')
  group.add_argument('-t', '--table', help='the name of the table whose schema should be displayed')
  return schema_parser


def _create_datasets_subparser(parser):
  datasets_parser = parser.subcommand('datasets', 'List the datasets in a BigQuery project.')
  datasets_parser.add_argument('-p', '--project',
                               help='The project whose datasets should be listed')
  datasets_parser.add_argument('-f', '--filter',
                               help='Optional wildcard filter string used to limit the results')
  return datasets_parser


def _create_tables_subparser(parser):
  tables_parser = parser.subcommand('tables', 'List the tables in a BigQuery project or dataset.')
  tables_parser.add_argument('-p', '--project',
                             help='The project whose tables should be listed')
  tables_parser.add_argument('-d', '--dataset',
                             help='The dataset to restrict to')
  tables_parser.add_argument('-f', '--filter',
                             help='Optional wildcard filter string used to limit the results')
  return tables_parser


def _create_extract_subparser(parser):
  extract_parser = parser.subcommand('extract', 'Extract BigQuery query results or table to GCS.')
  extract_parser.add_argument('-f', '--format', choices=['csv', 'json'], default='csv',
                              help='The format to use for the export')
  extract_parser.add_argument('-c', '--compress', action='store_true',
                              help='Whether to compress the data')
  extract_parser.add_argument('-H', '--header', action='store_true',
                              help='Whether to include a header line (CSV only)')
  extract_parser.add_argument('-d', '--delimiter', default=',',
                              help='The field delimiter to use (CSV only)')
  extract_parser.add_argument('-S', '--source', help='The name of the query or table to extract')
  extract_parser.add_argument('-D', '--destination', help='The URL of the destination')
  return extract_parser


def _create_load_subparser(parser):
  load_parser = parser.subcommand('load', 'Load data from GCS into a BigQuery table.')
  load_parser.add_argument('-m', '--mode', help='One of create (default), append or overwrite',
                           choices=['create', 'append', 'overwrite'], default='create')
  load_parser.add_argument('-f', '--format', help='The source format', choices=['json', 'csv'],
                           default='csv')
  load_parser.add_argument('-n', '--skip',
                           help='The number of initial lines to skip; useful for CSV headers',
                           type=int, default=0)
  load_parser.add_argument('-s', '--strict', help='Whether to reject bad values and jagged lines',
                           action='store_true')
  load_parser.add_argument('-d', '--delimiter', default=',',
                           help='The inter-field delimiter for CVS (default ,)')
  load_parser.add_argument('-q', '--quote', default='"',
                           help='The quoted field delimiter for CVS (default ")')
  load_parser.add_argument('-i', '--infer',
                           help='Whether to attempt to infer the schema from source; ' +
                               'if false the table must already exist',
                           action='store_true')
  load_parser.add_argument('-S', '--source', help='The URL of the GCS source(s)')
  load_parser.add_argument('-D', '--destination', help='The destination table name')
  return load_parser


def _get_query_argument(args, cell, env):
  """ Get a query argument to a cell magic.

  The query is specified with args['query']. We look that up and if it is a BQ query
  just return it. If it is instead a SqlModule or SqlStatement it may have variable
  references. We resolve those using the arg parser for the SqlModule, then override
  the resulting defaults with either the Python code in cell, or the dictionary in
  overrides. The latter is for if the overrides are specified with YAML or JSON and
  eventually we should eliminate code in favor of this.

  Args:
    args: the dictionary of magic arguments.
    cell: the cell contents which can be variable value overrides (if args has a 'query'
        value) or inline SQL otherwise.
    env: a dictionary that is used for looking up variable values.
  Returns:
    A Query object.
  """
  sql_arg = args.get('query', None)
  if sql_arg is None:
    # Assume we have inline SQL in the cell
    if not isinstance(cell, basestring):
      raise Exception('Expected a --query argument or inline SQL')
    return datalab.bigquery.Query(cell, values=env)

  item = datalab.utils.commands.get_notebook_item(sql_arg)
  if isinstance(item, datalab.bigquery.Query):  # Queries are already expanded.
    return item

  # Create an expanded BQ Query.
  config = datalab.utils.commands.parse_config(cell, env)
  item, env = datalab.data.SqlModule.get_sql_statement_with_environment(item, config)
  if cell:
    env.update(config)  # config is both a fallback and an override.
  return datalab.bigquery.Query(item, values=env)


def _sample_cell(args, cell_body):
  """Implements the bigquery sample cell magic for ipython notebooks.

  Args:
    args: the optional arguments following '%%bigquery sample'.
    cell_body: optional contents of the cell interpreted as SQL, YAML or JSON.
  Returns:
    The results of executing the sampling query, or a profile of the sample data.
  """

  env = datalab.utils.commands.notebook_environment()
  query = None
  table = None
  view = None

  if args['query']:
    query = _get_query_argument(args, cell_body, env)
  elif args['table']:
    table = _get_table(args['table'])
  elif args['view']:
    view = datalab.utils.commands.get_notebook_item(args['view'])
    if not isinstance(view, datalab.bigquery.View):
      raise Exception('%s is not a view' % args['view'])
  else:
    query = datalab.bigquery.Query(cell_body, values=env)

  count = args['count']
  method = args['method']
  if method == 'random':
    sampling = datalab.bigquery.Sampling.random(percent=args['percent'], count=count)
  elif method == 'hashed':
    sampling = datalab.bigquery.Sampling.hashed(field_name=args['field'],
                                            percent=args['percent'],
                                            count=count)
  elif method == 'sorted':
    ascending = args['order'] == 'ascending'
    sampling = datalab.bigquery.Sampling.sorted(args['field'],
                                            ascending=ascending,
                                            count=count)
  elif method == 'limit':
    sampling = datalab.bigquery.Sampling.default(count=count)
  else:
    sampling = datalab.bigquery.Sampling.default(count=count)

  if query:
    results = query.sample(sampling=sampling, dialect=args['dialect'],
                           billing_tier=args['billing'])
  elif view:
    results = view.sample(sampling=sampling)
  else:
    results = table.sample(sampling=sampling)
  if args['verbose']:
    print(results.sql)
  if args['profile']:
    return datalab.utils.commands.profile_df(results.to_dataframe())
  else:
    return results


def _create_cell(args, cell_body):
  """Implements the BigQuery cell magic used to create datasets and tables.

   The supported syntax is:

     %%bigquery create dataset -n|--name <name> [-f|--friendly <friendlyname>]
     [<description>]

   or:

     %%bigquery create table -n|--name <tablename> [--overwrite]
     [<YAML or JSON cell_body defining schema to use for tables>]

  Args:
    args: the argument following '%bigquery create <command>'.
  """
  if args['command'] == 'dataset':
    try:
      datalab.bigquery.Dataset(args['name']).create(friendly_name=args['friendly'],
                                                    description=cell_body)
    except Exception as e:
      print('Failed to create dataset %s: %s' % (args['name'], e))
  else:
    if cell_body is None:
      print('Failed to create %s: no schema specified' % args['name'])
    else:
      try:
        record = datalab.utils.commands.parse_config(cell_body, datalab.utils.commands.notebook_environment(), as_dict=False)
        schema = datalab.bigquery.Schema(record)
        datalab.bigquery.Table(args['name']).create(schema=schema, overwrite=args['overwrite'])
      except Exception as e:
        print('Failed to create table %s: %s' % (args['name'], e))


def _delete_cell(args, _):
  """Implements the BigQuery cell magic used to delete datasets and tables.

   The supported syntax is:

     %%bigquery delete dataset -n|--name <name>

   or:

     %%bigquery delete table -n|--name <name>

  Args:
    args: the argument following '%bigquery delete <command>'.
  """
  # TODO(gram): add support for wildchars and multiple arguments at some point. The latter is
  # easy, the former a bit more tricky if non-default projects are involved.
  if args['command'] == 'dataset':
    try:
      datalab.bigquery.Dataset(args['name']).delete()
    except Exception as e:
      print('Failed to delete dataset %s: %s' % (args['name'], e))
  else:
    try:
      datalab.bigquery.Table(args['name']).delete()
    except Exception as e:
      print('Failed to delete table %s: %s' % (args['name'], e))


def _dryrun_cell(args, cell_body):
  """Implements the BigQuery cell magic used to dry run BQ queries.

   The supported syntax is:
   %%bigquery dryrun [-q|--sql <query identifier>]
   [<YAML or JSON cell_body or inline SQL>]

  Args:
    args: the argument following '%bigquery dryrun'.
    cell_body: optional contents of the cell interpreted as YAML or JSON.
  Returns:
    The response wrapped in a DryRunStats object
  """
  query = _get_query_argument(args, cell_body, datalab.utils.commands.notebook_environment())

  if args['verbose']:
    print(query.sql)
  result = query.execute_dry_run(dialect=args['dialect'], billing_tier=args['billing'])
  return datalab.bigquery._query_stats.QueryStats(total_bytes=result['totalBytesProcessed'],
                                              is_cached=result['cacheHit'])


def _udf_cell(args, js):
  """Implements the bigquery_udf cell magic for ipython notebooks.

  The supported syntax is:
  %%bigquery udf --module <var>
  <js function>

  Args:
    args: the optional arguments following '%%bigquery udf'.
    js: the UDF declaration (inputs and outputs) and implementation in javascript.
  Returns:
    The results of executing the UDF converted to a dataframe if no variable
    was specified. None otherwise.
  """
  variable_name = args['module']
  if not variable_name:
    raise Exception('Declaration must be of the form %%bigquery udf --module <variable name>')

  # Parse out the input and output specification
  spec_pattern = r'\{\{([^}]+)\}\}'
  spec_part_pattern = r'[a-z_][a-z0-9_]*'

  specs = re.findall(spec_pattern, js)
  if len(specs) < 2:
    raise Exception('The JavaScript must declare the input row and output emitter parameters '
                    'using valid jsdoc format comments.\n'
                    'The input row param declaration must be typed as {{field:type, field2:type}} '
                    'and the output emitter param declaration must be typed as '
                    'function({{field:type, field2:type}}.')

  inputs = []
  input_spec_parts = re.findall(spec_part_pattern, specs[0], flags=re.IGNORECASE)
  if len(input_spec_parts) % 2 != 0:
    raise Exception('Invalid input row param declaration. The jsdoc type expression must '
                    'define an object with field and type pairs.')
  for n, t in zip(input_spec_parts[0::2], input_spec_parts[1::2]):
    inputs.append((n, t))

  outputs = []
  output_spec_parts = re.findall(spec_part_pattern, specs[1], flags=re.IGNORECASE)
  if len(output_spec_parts) % 2 != 0:
    raise Exception('Invalid output emitter param declaration. The jsdoc type expression must '
                    'define a function accepting an an object with field and type pairs.')
  for n, t in zip(output_spec_parts[0::2], output_spec_parts[1::2]):
    outputs.append((n, t))

  # Look for imports. We use a non-standard @import keyword; we could alternatively use @requires.
  # Object names can contain any characters except \r and \n.
  import_pattern = r'@import[\s]+(gs://[a-z\d][a-z\d_\.\-]*[a-z\d]/[^\n\r]+)'
  imports = re.findall(import_pattern, js)

  # Split the cell if necessary. We look for a 'function(' with no name and a header comment
  # block with @param and assume this is the primary function, up to a closing '}' at the start
  # of the line. The remaining cell content is used as support code.
  split_pattern = r'(.*)(/\*.*?@param.*?@param.*?\*/\w*\n\w*function\w*\(.*?^}\n?)(.*)'
  parts = re.match(split_pattern, js, re.MULTILINE | re.DOTALL)
  support_code = ''
  if parts:
    support_code = (parts.group(1) + parts.group(3)).strip()
    if len(support_code):
      js = parts.group(2)

  # Finally build the UDF object
  udf = datalab.bigquery.UDF(inputs, outputs, variable_name, js, support_code, imports)
  datalab.utils.commands.notebook_environment()[variable_name] = udf


def _execute_cell(args, cell_body):
  """Implements the BigQuery cell magic used to execute BQ queries.

   The supported syntax is:
   %%bigquery execute [-q|--sql <query identifier>] <other args>
   [<YAML or JSON cell_body or inline SQL>]

  Args:
    args: the arguments following '%bigquery execute'.
    cell_body: optional contents of the cell interpreted as YAML or JSON.
  Returns:
    The QueryResultsTable
  """
  query = _get_query_argument(args, cell_body, datalab.utils.commands.notebook_environment())
  if args['verbose']:
    print(query.sql)
  return query.execute(args['target'], table_mode=args['mode'], use_cache=not args['nocache'],
                       allow_large_results=args['large'], dialect=args['dialect'],
                       billing_tier=args['billing']).results


def _pipeline_cell(args, cell_body):
  """Implements the BigQuery cell magic used to validate, execute or deploy BQ pipelines.

   The supported syntax is:
   %%bigquery pipeline [-q|--sql <query identifier>] <other args> <action>
   [<YAML or JSON cell_body or inline SQL>]

  Args:
    args: the arguments following '%bigquery pipeline'.
    cell_body: optional contents of the cell interpreted as YAML or JSON.
  Returns:
    The QueryResultsTable
  """
  if args['action'] == 'deploy':
    raise Exception('Deploying a pipeline is not yet supported')

  env = {}
  for key, value in datalab.utils.commands.notebook_environment().items():
    if isinstance(value, datalab.bigquery._udf.UDF):
      env[key] = value

  query = _get_query_argument(args, cell_body, env)
  if args['verbose']:
    print(query.sql)
  if args['action'] == 'dryrun':
    print(query.sql)
    result = query.execute_dry_run()
    return datalab.bigquery._query_stats.QueryStats(total_bytes=result['totalBytesProcessed'],
                                                is_cached=result['cacheHit'])
  if args['action'] == 'run':
    return query.execute(args['target'], table_mode=args['mode'], use_cache=not args['nocache'],
                         allow_large_results=args['large'], dialect=args['dialect'],
                         billing_tier=args['billing']).results


def _table_line(args):
  """Implements the BigQuery table magic used to display tables.

   The supported syntax is:
   %bigquery table -t|--table <name> <other args>

  Args:
    args: the arguments following '%bigquery table'.
  Returns:
    The HTML rendering for the table.
  """
  # TODO(gram): It would be good to turn _table_viewer into a class that has a registered
  # renderer. That would allow this to return a table viewer object which is easier to test.
  name = args['table']
  table = _get_table(name)
  if table and table.exists():
    fields = args['cols'].split(',') if args['cols'] else None
    html = _table_viewer(table, rows_per_page=args['rows'], fields=fields)
    return IPython.core.display.HTML(html)
  else:
    raise Exception('Table %s does not exist; cannot display' % name)


def _get_schema(name):
  """ Given a variable or table name, get the Schema if it exists. """
  item = datalab.utils.commands.get_notebook_item(name)
  if not item:
    item = _get_table(name)

  if isinstance(item, datalab.bigquery.Schema):
    return item
  if hasattr(item, 'schema') and isinstance(item.schema, datalab.bigquery._schema.Schema):
    return item.schema
  return None


# An LRU cache for Tables. This is mostly useful so that when we cross page boundaries
# when paging through a table we don't have to re-fetch the schema.
_table_cache = datalab.utils.LRUCache(10)


def _get_table(name):
  """ Given a variable or table name, get a Table if it exists.

  Args:
    name: the name of the Table or a variable referencing the Table.
  Returns:
    The Table, if found.
  """
  # If name is a variable referencing a table, use that.
  item = datalab.utils.commands.get_notebook_item(name)
  if isinstance(item, datalab.bigquery.Table):
    return item
  # Else treat this as a BQ table name and return the (cached) table if it exists.
  try:
    return _table_cache[name]
  except KeyError:
    table = datalab.bigquery.Table(name)
    if table.exists():
      _table_cache[name] = table
      return table
  return None


def _schema_line(args):
  """Implements the BigQuery schema magic used to display table/view schemas.

  Args:
    args: the arguments following '%bigquery schema'.
  Returns:
    The HTML rendering for the schema.
  """
  # TODO(gram): surely we could just return the schema itself?
  name = args['table'] if args['table'] else args['view']
  if name is None:
    raise Exception('No table or view specified; cannot show schema')

  schema = _get_schema(name)
  if schema:
    html = _repr_html_table_schema(schema)
    return IPython.core.display.HTML(html)
  else:
    raise Exception('%s is not a schema and does not appear to have a schema member' % name)


def _render_table(data, fields=None):
  """ Helper to render a list of dictionaries as an HTML display object. """
  return IPython.core.display.HTML(datalab.utils.commands.HtmlBuilder.render_table(data, fields))


def _render_list(data):
  """ Helper to render a list of objects as an HTML list object. """
  return IPython.core.display.HTML(datalab.utils.commands.HtmlBuilder.render_list(data))


def _datasets_line(args):
  """Implements the BigQuery datasets magic used to display datasets in a project.

   The supported syntax is:

       %bigquery datasets [-f <filter>] [-p|--project <project_id>]

  Args:
    args: the arguments following '%bigquery datasets'.
  Returns:
    The HTML rendering for the table of datasets.
  """
  filter_ = args['filter'] if args['filter'] else '*'
  return _render_list([str(dataset) for dataset in datalab.bigquery.Datasets(args['project'])
                       if fnmatch.fnmatch(str(dataset), filter_)])


def _tables_line(args):
  """Implements the BigQuery tables magic used to display tables in a dataset.

   The supported syntax is:

       %bigquery tables -p|--project <project_id>  -d|--dataset <dataset_id>

  Args:
    args: the arguments following '%bigquery tables'.
  Returns:
    The HTML rendering for the list of tables.
  """
  filter_ = args['filter'] if args['filter'] else '*'
  if args['dataset']:
    if args['project'] is None:
      datasets = [datalab.bigquery.Dataset(args['dataset'])]
    else:
      datasets = [datalab.bigquery.Dataset((args['project'], args['dataset']))]
  else:
    datasets = datalab.bigquery.Datasets(args['project'])

  tables = []
  for dataset in datasets:
    tables.extend([str(table) for table in dataset if fnmatch.fnmatch(str(table), filter_)])

  return _render_list(tables)


def _extract_line(args):
  """Implements the BigQuery extract magic used to extract table data to GCS.

   The supported syntax is:

       %bigquery extract -S|--source <table> -D|--destination <url> <other_args>

  Args:
    args: the arguments following '%bigquery extract'.
  Returns:
    A message about whether the extract succeeded or failed.
  """
  name = args['source']
  source = datalab.utils.commands.get_notebook_item(name)
  if not source:
    source = _get_table(name)

  if not source:
    raise Exception('No source named %s found' % name)
  elif isinstance(source, datalab.bigquery.Table) and not source.exists():
    raise Exception('Table %s does not exist' % name)
  else:

    job = source.extract(args['destination'],
                         format='CSV' if args['format'] == 'csv' else 'NEWLINE_DELIMITED_JSON',
                         compress=args['compress'],
                         csv_delimiter=args['delimiter'],
                         csv_header=args['header'])
    if job.failed:
      raise Exception('Extract failed: %s' % str(job.fatal_error))
    elif job.errors:
      raise Exception('Extract completed with errors: %s' % str(job.errors))


def _load_cell(args, schema):
  """Implements the BigQuery load magic used to load data from GCS to a table.

   The supported syntax is:

       %bigquery load -S|--source <source> -D|--destination <table>  <other_args>

  Args:
    args: the arguments following '%bigquery load'.
    schema: a JSON schema for the destination table.
  Returns:
    A message about whether the load succeeded or failed.
  """
  name = args['destination']
  table = _get_table(name)
  if not table:
    table = datalab.bigquery.Table(name)

  if table.exists():
    if args['mode'] == 'create':
      raise Exception('%s already exists; use --append or --overwrite' % name)
  elif schema:
    table.create(json.loads(schema))
  elif not args['infer']:
    raise Exception(
        'Table does not exist, no schema specified in cell and no --infer flag; cannot load')

  # TODO(gram): we should probably try do the schema infer ourselves as BQ doesn't really seem
  # to be able to do it. Alternatively we can drop the --infer argument and force the user
  # to use a pre-existing table or supply a JSON schema.
  csv_options = datalab.bigquery.CSVOptions(delimiter=args['delimiter'],
                                        skip_leading_rows=args['skip'],
                                        allow_jagged_rows=not args['strict'],
                                        quote=args['quote'])
  job = table.load(args['source'],
                   mode=args['mode'],
                   source_format=('CSV' if args['format'] == 'csv' else 'NEWLINE_DELIMITED_JSON'),
                   csv_options=csv_options,
                   ignore_unknown_values=not args['strict'])
  if job.failed:
    raise Exception('Load failed: %s' % str(job.fatal_error))
  elif job.errors:
    raise Exception('Load completed with errors: %s' % str(job.errors))


def _add_command(parser, subparser_fn, handler, cell_required=False, cell_prohibited=False):
  """ Create and initialize a bigquery subcommand handler. """
  sub_parser = subparser_fn(parser)
  sub_parser.set_defaults(func=lambda args, cell: _dispatch_handler(args, cell, sub_parser, handler,
                          cell_required=cell_required, cell_prohibited=cell_prohibited))


def _create_bigquery_parser():
  """ Create the parser for the %bigquery magics.

  Note that because we use the func default handler dispatch mechanism of argparse,
  our handlers can take only one argument which is the parsed args. So we must create closures
  for the handlers that bind the cell contents and thus must recreate this parser for each
  cell upon execution.
  """
  parser = datalab.utils.commands.CommandParser(prog='bigquery', description="""
Execute various BigQuery-related operations. Use "%bigquery <command> -h"
for help on a specific command.
  """)

  # This is a bit kludgy because we want to handle some line magics and some cell magics
  # with the bigquery command.

  # %%bigquery sample
  _add_command(parser, _create_sample_subparser, _sample_cell)

  # %%bigquery create
  _add_command(parser, _create_create_subparser, _create_cell)

  # %%bigquery delete
  _add_command(parser, _create_delete_subparser, _delete_cell)

  # %%bigquery dryrun
  _add_command(parser, _create_dry_run_subparser, _dryrun_cell)

  # %%bigquery udf
  _add_command(parser, _create_udf_subparser, _udf_cell, cell_required=True)

  # %%bigquery execute
  _add_command(parser, _create_execute_subparser, _execute_cell)

  # %%bigquery pipeline
  _add_command(parser, _create_pipeline_subparser, _pipeline_cell)

  # %bigquery table
  _add_command(parser, _create_table_subparser, _table_line, cell_prohibited=True)

  # %bigquery schema
  _add_command(parser, _create_schema_subparser, _schema_line, cell_prohibited=True)

  # %bigquery datasets
  _add_command(parser, _create_datasets_subparser, _datasets_line, cell_prohibited=True)

  # %bigquery tables
  _add_command(parser, _create_tables_subparser, _tables_line, cell_prohibited=True)

  # % bigquery extract
  _add_command(parser, _create_extract_subparser, _extract_line, cell_prohibited=True)

  # %bigquery load
  # TODO(gram): need some additional help, esp. around the option of specifying schema in
  # cell body and how schema infer may fail.
  _add_command(parser, _create_load_subparser, _load_cell)
  return parser


_bigquery_parser = _create_bigquery_parser()


@IPython.core.magic.register_line_cell_magic
def bigquery(line, cell=None):
  """Implements the bigquery cell magic for ipython notebooks.

  The supported syntax is:

    %%bigquery <command> [<args>]
    <cell>

  or:

    %bigquery <command> [<args>]

  Use %bigquery --help for a list of commands, or %bigquery <command> --help for help
  on a specific command.
  """
  namespace = {}
  if line.find('$') >= 0:
    # We likely have variables to expand; get the appropriate context.
    namespace = datalab.utils.commands.notebook_environment()

  return datalab.utils.commands.handle_magic_line(line, cell, _bigquery_parser, namespace=namespace)


def _dispatch_handler(args, cell, parser, handler, cell_required=False, cell_prohibited=False):
  """ Makes sure cell magics include cell and line magics don't, before dispatching to handler.

  Args:
    args: the parsed arguments from the magic line.
    cell: the contents of the cell, if any.
    parser: the argument parser for <cmd>; used for error message.
    handler: the handler to call if the cell present/absent check passes.
    cell_required: True for cell magics, False for line magics that can't be cell magics.
    cell_prohibited: True for line magics, False for cell magics that can't be line magics.
  Returns:
    The result of calling the handler.
  Raises:
    Exception if the invocation is not valid.
  """
  if cell_prohibited:
    if cell and len(cell.strip()):
      parser.print_help()
      raise Exception('Additional data is not supported with the %s command.' % parser.prog)
    return handler(args)

  if cell_required and not cell:
    parser.print_help()
    raise Exception('The %s command requires additional data' % parser.prog)

  return handler(args, cell)


def _table_viewer(table, rows_per_page=25, fields=None):
  """  Return a table viewer.

    This includes a static rendering of the first page of the table, that gets replaced
    by the charting code in environments where Javascript is executable and BQ is available.

  Args:
    table: the table to view.
    rows_per_page: how many rows to display at one time.
    fields: an array of field names to display; default is None which uses the full schema.
  Returns:
    A string containing the HTML for the table viewer.
  """

  # TODO(gram): rework this to use datalab.utils.commands.chart_html

  if not table.exists():
    raise Exception('Table %s does not exist' % str(table))

  _HTML_TEMPLATE = u"""
    <div class="bqtv" id="{div_id}">{static_table}</div>
    <br />{meta_data}<br />
    <script>

      require.config({{
        paths: {{
          d3: '//cdnjs.cloudflare.com/ajax/libs/d3/3.4.13/d3',
          plotly: 'https://cdn.plot.ly/plotly-1.5.1.min.js?noext',
          jquery: '//ajax.googleapis.com/ajax/libs/jquery/2.0.0/jquery.min'
        }},
        map: {{
          '*': {{
            datalab: 'nbextensions/gcpdatalab'
          }}
        }},
        shim: {{
          plotly: {{
            deps: ['d3', 'jquery'],
            exports: 'plotly'
          }}
        }}
      }});

      require(['datalab/charting', 'datalab/element!{div_id}', 'base/js/events',
          'datalab/style!/nbextensions/gcpdatalab/charting.css'],
        function(charts, dom, events) {{
          charts.render('gcharts', dom, events, '{chart_style}', [], {data},
            {{
              pageSize: {rows_per_page},
              cssClassNames:  {{
                tableRow: 'gchart-table-row',
                headerRow: 'gchart-table-headerrow',
                oddTableRow: 'gchart-table-oddrow',
                selectedTableRow: 'gchart-table-selectedrow',
                hoverTableRow: 'gchart-table-hoverrow',
                tableCell: 'gchart-table-cell',
                headerCell: 'gchart-table-headercell',
                rowNumberCell: 'gchart-table-rownumcell'
              }}
            }},
            {{source_index: {source_index}, fields: '{fields}'}},
            0,
            {total_rows});
        }}
      );
    </script>
  """

  if fields is None:
    fields = datalab.utils.commands.get_field_list(fields, table.schema)
  div_id = datalab.utils.commands.Html.next_id()
  meta_count = ('rows: %d' % table.length) if table.length >= 0 else ''
  meta_name = str(table) if table.job is None else ('job: %s' % table.job.id)
  if table.job:
    if table.job.cache_hit:
      meta_cost = 'cached'
    else:
      bytes = datalab.bigquery._query_stats.QueryStats._size_formatter(table.job.bytes_processed)
      meta_cost = '%s processed' % bytes
    meta_time = 'time: %.1fs' % table.job.total_time
  else:
    meta_cost = ''
    meta_time = ''

  data, total_count = datalab.utils.commands.get_data(table, fields, first_row=0, count=rows_per_page)

  if total_count < 0:
    # The table doesn't have a length metadata property but may still be small if we fetched less
    # rows than we asked for.
    fetched_count = len(data['rows'])
    if fetched_count < rows_per_page:
      total_count = fetched_count

  chart = 'table' if 0 <= total_count <= rows_per_page else 'paged_table'
  meta_entries = [meta_count, meta_time, meta_cost, meta_name]
  meta_data = '(%s)' % (', '.join([entry for entry in meta_entries if len(entry)]))

  return _HTML_TEMPLATE.format(div_id=div_id,
                               static_table=datalab.utils.commands.HtmlBuilder.render_chart_data(data),
                               meta_data=meta_data,
                               chart_style=chart,
                               source_index=datalab.utils.commands.get_data_source_index(str(table)),
                               fields=','.join(fields),
                               total_rows=total_count,
                               rows_per_page=rows_per_page,
                               data=json.dumps(data, cls=datalab.utils.JSONEncoder))


def _repr_html_query(query):
  # TODO(nikhilko): Pretty print the SQL
  return datalab.utils.commands.HtmlBuilder.render_text(query.sql, preformatted=True)


def _repr_html_query_results_table(results):
  return _table_viewer(results)


def _repr_html_table(results):
  return _table_viewer(results)


def _repr_html_table_schema(schema):
  _HTML_TEMPLATE = """
    <div class="bqsv" id="%s"></div>
    <script>
      require.config({
        map: {
          '*': {
            datalab: 'nbextensions/gcpdatalab'
          }
        },
      });

      require(['datalab/bigquery', 'datalab/element!%s',
          'datalab/style!/nbextensions/gcpdatalab/bigquery.css'],
        function(bq, dom) {
          bq.renderSchema(dom, %s);
        }
      );
    </script>
    """
  id = datalab.utils.commands.Html.next_id()
  return _HTML_TEMPLATE % (id, id, json.dumps(schema._bq_schema))


def _register_html_formatters():
  try:
    ipy = IPython.get_ipython()
    html_formatter = ipy.display_formatter.formatters['text/html']

    html_formatter.for_type_by_name('datalab.bigquery._query', 'Query', _repr_html_query)
    html_formatter.for_type_by_name('datalab.bigquery._query_results_table', 'QueryResultsTable',
                                    _repr_html_query_results_table)
    html_formatter.for_type_by_name('datalab.bigquery._table', 'Table', _repr_html_table)
    html_formatter.for_type_by_name('datalab.bigquery._schema', 'Schema', _repr_html_table_schema)
  except TypeError:
    # For when running unit tests
    pass

_register_html_formatters()



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Useful common utility functions."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import str
from past.builtins import basestring

import collections
import re


DatasetName = collections.namedtuple('DatasetName', ['project_id', 'dataset_id'])
""" A namedtuple for Dataset names.

  Args:
    project_id: the project id for the dataset.
    dataset_id: the dataset id for the dataset.
"""

TableName = collections.namedtuple('TableName',
                                   ['project_id', 'dataset_id', 'table_id', 'decorator'])
""" A namedtuple for Table names.

  Args:
    project_id: the project id for the table.
    dataset_id: the dataset id for the table.
    table_id: the table id for the table.
    decorator: the optional decorator for the table (for windowing/snapshot-ing).
"""

# Absolute project-qualified name pattern: <project>:<dataset>
_ABS_DATASET_NAME_PATTERN = r'^([a-z\d\-_\.:]+)\:(\w+)$'

# Relative name pattern: <dataset>
_REL_DATASET_NAME_PATTERN = r'^(\w+)$'

# Absolute project-qualified name pattern: <project>:<dataset>.<table>
_ABS_TABLE_NAME_PATTERN = r'^([a-z\d\-_\.:]+)\:(\w+)\.(\w+)(@[\d\-]+)?$'

# Relative name pattern: <dataset>.<table>
_REL_TABLE_NAME_PATTERN = r'^(\w+)\.(\w+)(@[\d\-]+)?$'

# Table-only name pattern: <table>. Includes an optional decorator.
_TABLE_NAME_PATTERN = r'^(\w+)(@[\d\-]+)$'


def parse_dataset_name(name, project_id=None):
  """Parses a dataset name into its individual parts.

  Args:
    name: the name to parse, or a tuple, dictionary or array containing the parts.
    project_id: the expected project ID. If the name does not contain a project ID,
        this will be used; if the name does contain a project ID and it does not match
        this, an exception will be thrown.
  Returns:
    A DatasetName named tuple for the dataset.
  Raises:
    Exception: raised if the name doesn't match the expected formats or a project_id was
        specified that does not match that in the name.
  """
  _project_id = _dataset_id = None
  if isinstance(name, basestring):
    # Try to parse as absolute name first.
    m = re.match(_ABS_DATASET_NAME_PATTERN, name, re.IGNORECASE)
    if m is not None:
      _project_id, _dataset_id = m.groups()
    else:
      # Next try to match as a relative name implicitly scoped within current project.
      m = re.match(_REL_DATASET_NAME_PATTERN, name)
      if m is not None:
        groups = m.groups()
        _dataset_id = groups[0]
  elif isinstance(name, dict):
    try:
      _dataset_id = name['dataset_id']
      _project_id = name['project_id']
    except KeyError:
      pass
  else:
    # Try treat as an array or tuple
    if len(name) == 2:
      # Treat as a tuple or array.
      _project_id, _dataset_id = name
    elif len(name) == 1:
      _dataset_id = name[0]
  if not _dataset_id:
    raise Exception('Invalid dataset name: ' + str(name))
  if not _project_id:
    _project_id = project_id

  return DatasetName(_project_id, _dataset_id)


def parse_table_name(name, project_id=None, dataset_id=None):
  """Parses a table name into its individual parts.

  Args:
    name: the name to parse, or a tuple, dictionary or array containing the parts.
    project_id: the expected project ID. If the name does not contain a project ID,
        this will be used; if the name does contain a project ID and it does not match
        this, an exception will be thrown.
    dataset_id: the expected dataset ID. If the name does not contain a dataset ID,
        this will be used; if the name does contain a dataset ID and it does not match
        this, an exception will be thrown.
  Returns:
    A TableName named tuple consisting of the full name and individual name parts.
  Raises:
    Exception: raised if the name doesn't match the expected formats, or a project_id and/or
        dataset_id was provided that does not match that in the name.
  """
  _project_id = _dataset_id = _table_id = _decorator = None
  if isinstance(name, basestring):
    # Try to parse as absolute name first.
    m = re.match(_ABS_TABLE_NAME_PATTERN, name, re.IGNORECASE)
    if m is not None:
      _project_id, _dataset_id, _table_id, _decorator = m.groups()
    else:
      # Next try to match as a relative name implicitly scoped within current project.
      m = re.match(_REL_TABLE_NAME_PATTERN, name)
      if m is not None:
        groups = m.groups()
        _project_id, _dataset_id, _table_id, _decorator =\
            project_id, groups[0], groups[1], groups[2]
      else:
        # Finally try to match as a table name only.
        m = re.match(_TABLE_NAME_PATTERN, name)
        if m is not None:
          groups = m.groups()
          _project_id, _dataset_id, _table_id, _decorator =\
              project_id, dataset_id, groups[0], groups[1]
  elif isinstance(name, dict):
    try:
      _table_id = name['table_id']
      _dataset_id = name['dataset_id']
      _project_id = name['project_id']
    except KeyError:
      pass
  else:
    # Try treat as an array or tuple
    if len(name) == 4:
      _project_id, _dataset_id, _table_id, _decorator = name
    elif len(name) == 3:
      _project_id, _dataset_id, _table_id = name
    elif len(name) == 2:
      _dataset_id, _table_id = name
  if not _table_id:
    raise Exception('Invalid table name: ' + str(name))
  if not _project_id:
    _project_id = project_id
  if not _dataset_id:
    _dataset_id = dataset_id
  if not _decorator:
    _decorator = ''

  return TableName(_project_id, _dataset_id, _table_id, _decorator)


def format_query_errors(errors):
  return '\n'.join(['%s: %s' % (error['reason'], error['message']) for error in errors])





# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations under
# the License.

"""Provides utility methods for the Monitoring API."""

from __future__ import absolute_import

import google.cloud.monitoring

import datalab.context


def make_client(project_id=None, context=None):
  context = context or datalab.context.Context.default()
  project_id = project_id or context.project_id
  client = google.cloud.monitoring.Client(
      project=project_id,
      credentials=context.credentials,
  )
  client._connection_class.USER_AGENT = 'pydatalab/v0'
  return client



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""A simple LRU cache."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import str
from past.builtins import basestring
from builtins import object

import datetime


class LRUCache(object):
  """A simple LRU cache."""

  def __init__(self, cache_size):
    """ Initialize the cache with the given size.

    Args:
      cache_size: the maximum number of items the cache can hold. Attempts to add more
          items than this will result in the least recently used items being displaced to
          make room.
    """
    self._cache = {}
    self._cache_size = cache_size

  def __getitem__(self, key):
    """ Get an item from the cache.

    Args:
      key: a string used as the lookup key.
    Returns:
      The cached item, if any.
    Raises:
      Exception if the key is not a string.
      KeyError if the key is not found.
    """
    if not isinstance(key, basestring):
      raise Exception("LRU cache can only be indexed by strings (%s has type %s)" %
                      (str(key), str(type(key))))

    if key in self._cache:
      entry = self._cache[key]
      entry['last_used'] = datetime.datetime.now()
      return entry['value']
    else:
      raise KeyError(key)

  def __delitem__(self, key):
    """ Remove an item from the cache.

    Args:
      key: a string key for retrieving the item.
    """
    if not isinstance(key, basestring):
      raise Exception("LRU cache can only be indexed by strings")
    del self._cache[key]

  def __setitem__(self, key, value):
    """ Put an item in the cache.

    Args:
      key: a string key for retrieving the item.
      value: the item to cache.
    Raises:
      Exception if the key is not a string.
    """
    if not isinstance(key, basestring):
      raise Exception("LRU cache can only be indexed by strings")

    if key in self._cache:
      entry = self._cache[key]
    elif len(self._cache) < self._cache_size:
      # Cache is not full; append an new entry
      self._cache[key] = entry = {}
    else:
      # Cache is full; displace an entry
      entry = min(list(self._cache.values()), key=lambda x: x['last_used'])
      self._cache.pop(entry['key'])
      self._cache[key] = entry

    entry['value'] = value
    entry['key'] = key
    entry['last_used'] = datetime.datetime.now()

  def __contains__(self, key):
    return key in self._cache

  def get(self, key, value):
    if key in self._cache:
      return self._cache[key]['value']
    return value




# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.


from plotly.offline import iplot


class ConfusionMatrix(object):
  """Represents a confusion matrix."""

  def __init__(self, predicted_labels, true_labels, counts):
    """Initializes an instance of a ComfusionMatrix. the length of predicted_values,
       true_values, count must be the same.

    Args:
      predicted_labels: a list of predicted labels.
      true_labels: a list of true labels.
      counts: a list of count for each (predicted, true) combination.

    Raises: Exception if predicted_labels, true_labels, and counts are not of the same size
    """
    if len(predicted_labels) != len(true_labels) or len(true_labels) != len(counts):
      raise Exception('The input predicted_labels, true_labels, counts need to be same size.')
    self._all_labels = list(set(predicted_labels) | set(true_labels))
    data = []
    for value in self._all_labels:
      predicts_for_current_true_label = \
          {p: c for p, t, c in zip(predicted_labels, true_labels, counts) if t == value}
      # sort by all_values and fill in zeros if needed
      predicts_for_current_true_label = [predicts_for_current_true_label.get(v, 0)
          for v in self._all_labels]
      data.append(predicts_for_current_true_label)
    self._data = data

  def plot(self):
    """Plot the confusion matrix."""
    figure_data = \
    {
      "data": [
        {
          "x": self._all_labels,
          "y": self._all_labels,
          "z": self._data,
          "colorscale": "YlGnBu",
          "type": "heatmap"
        }
      ],
      "layout": {
        "title": "Confusion Matrix",
        "xaxis": {
          "title": "Predicted value",
        },
        "yaxis": {
          "title": "True Value",
        }
      }
    }
    iplot(figure_data)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements External Table functionality."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import object

from . import _csv_options


class FederatedTable(object):

  @staticmethod
  def from_storage(source, source_format='csv', csv_options=None, ignore_unknown_values=False,
                   max_bad_records=0, compressed=False, schema=None):

    """ Create an external table for a GCS object.

    Args:
      source: the URL of the source objects(s). Can include a wildcard '*' at the end of the item
         name. Can be a single source or a list.
      source_format: the format of the data, 'csv' or 'json'; default 'csv'.
      csv_options: For CSV files, the options such as quote character and delimiter.
      ignore_unknown_values: If True, accept rows that contain values that do not match the schema;
          the unknown values are ignored (default False).
      max_bad_records: The maximum number of bad records that are allowed (and ignored) before
          returning an 'invalid' error in the Job result (default 0).
      compressed: whether the data is GZ compressed or not (default False). Note that compressed
          data can be used as a federated table but cannot be loaded into a BQ Table.
      schema: the schema of the data. This is required for this table to be used as a federated
          table or to be loaded using a Table object that itself has no schema (default None).

  """
    result = FederatedTable()
    # Do some sanity checking and concert some params from friendly form to form used by BQ.
    if source_format == 'csv':
      result._bq_source_format = 'CSV'
      if csv_options is None:
        csv_options = _csv_options.CSVOptions()  # use defaults
    elif source_format == 'json':
      if csv_options:
        raise Exception('CSV options are not support for JSON tables')
      result._bq_source_format = 'NEWLINE_DELIMITED_JSON'
    else:
      raise Exception("Invalid source format %s" % source_format)

    result._source = source if isinstance(source, list) else [source]
    result._source_format = source_format
    result._csv_options = csv_options
    result._ignore_unknown_values = ignore_unknown_values
    result._max_bad_records = max_bad_records
    result._compressed = compressed
    result._schema = schema
    return result

  def __init__(self):

    """ Create an external table reference. Do not call this directly; use factory method(s). """
    # Do some sanity checking and concert some params from friendly form to form used by BQ.
    self._bq_source_format = None
    self._source = None
    self._source_format = None
    self._csv_options = None
    self._ignore_unknown_values = None
    self._max_bad_records = None
    self._compressed = None
    self._schema = None

  @property
  def schema(self):
    return self._schema

  def _to_query_json(self):
    """ Return the table as a dictionary to be used as JSON in a query job. """
    json = {
      'compression': 'GZIP' if self._compressed else 'NONE',
      'ignoreUnknownValues': self._ignore_unknown_values,
      'maxBadRecords': self._max_bad_records,
      'sourceFormat': self._bq_source_format,
      'sourceUris': self._source,
    }
    if self._source_format == 'csv' and self._csv_options:
      json['csvOptions'] = {}
      json['csvOptions'].update(self._csv_options._to_query_json())
    if self._schema:
      json['schema'] = {'fields': self._schema._bq_schema}
    return json




# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements Job functionality for async tasks."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import str
from builtins import object

import concurrent.futures
import datetime
import time
import traceback
import uuid


class JobError(Exception):
  """ A helper class to capture multiple components of Job errors.  """

  def __init__(self, location, message, reason):
    self.location = location
    self.message = message
    self.reason = reason

  def __str__(self):
    return self.message


class Job(object):
  """A manager object for async operations.

     A Job can have a Future in which case it will be able to monitor its own completion state
     and result, or it may have no Future in which case it must be a derived class that
     manages this some other way. We do this instead of having an abstract base class in
     order to make wait_one/wait_all more efficient; instead of just sleeping and polling
     we can use more reactive ways of monitoring groups of Jobs.
  """

  _POLL_INTERVAL_SECONDS = 5

  def __init__(self, job_id=None, future=None):
    """Initializes an instance of a Job.

    Args:
      job_id: a unique ID for the job. If None, a UUID will be generated.
      future: the Future associated with the Job, if any.
    """
    self._job_id = str(uuid.uuid4()) if job_id is None else job_id
    self._future = future
    self._is_complete = False
    self._errors = None
    self._fatal_error = None
    self._result = None
    self._start_time = datetime.datetime.utcnow()
    self._end_time = None

  def __str__(self):
    return self._job_id

  @property
  def id(self):
    """ Get the Job ID.

    Returns:
      The ID of the job.
    """
    return self._job_id

  @property
  def is_complete(self):
    """ Get the completion state of the job.

    Returns:
      True if the job is complete; False if it is still running.
    """
    self._refresh_state()
    return self._is_complete

  @property
  def failed(self):
    """ Get the success state of the job.

    Returns:
      True if the job failed; False if it is still running or succeeded (possibly with partial
      failure).
    """
    self._refresh_state()
    return self._is_complete and self._fatal_error is not None

  @property
  def fatal_error(self):
    """ Get the job error.

    Returns:
      None if the job succeeded or is still running, else the error tuple for the failure.
    """
    self._refresh_state()
    return self._fatal_error

  @property
  def errors(self):
    """ Get the non-fatal errors in the job.

    Returns:
      None if the job is still running, else the list of errors that occurred.
    """
    self._refresh_state()
    return self._errors

  def result(self):
    """ Get the result for a job. This will block if the job is incomplete.

    Returns:
      The result for the Job.

    Raises:
      An exception if the Job resulted in an exception.

    """
    self.wait()
    if self._fatal_error:
      raise self._fatal_error
    return self._result

  @property
  def start_time_utc(self):
    """ The UTC start time of the job as a Python datetime. """
    return self._start_time

  @property
  def end_time_utc(self):
    """ The UTC end time of the job (or None if incomplete) as a Python datetime. """
    return self._end_time

  @property
  def total_time(self):
    """ The total time in fractional seconds that the job took, or None if not complete. """
    if self._end_time is None:
      return None
    return (self._end_time - self._start_time).total_seconds()

  def _refresh_state(self):
    """ Get the state of a job. Must be overridden by derived Job classes
        for Jobs that don't use a Future.
    """
    if self._is_complete:
      return

    if not self._future:
      raise Exception('Please implement this in the derived class')

    if self._future.done():
      self._is_complete = True
      self._end_time = datetime.datetime.utcnow()
      try:
        self._result = self._future.result()
      except Exception as e:
        message = str(e)
        self._fatal_error = JobError(location=traceback.format_exc(), message=message,
                                     reason=str(type(e)))

  def _timeout(self):
    """ Helper for raising timeout errors. """
    raise concurrent.futures.TimeoutError('Timed out waiting for Job %s to complete' % self._job_id)

  def wait(self, timeout=None):
    """ Wait for the job to complete, or a timeout to happen.

    Args:
      timeout: how long to wait before giving up (in seconds); default None which means no timeout.

    Returns:
      The Job
    """
    if self._future:
      try:
        # Future.exception() will return rather than raise any exception so we use it.
        self._future.exception(timeout)
      except concurrent.futures.TimeoutError:
        self._timeout()
      self._refresh_state()
    else:
      # fall back to polling
      while not self.is_complete:
        if timeout is not None:
          if timeout <= 0:
            self._timeout()
          timeout -= Job._POLL_INTERVAL_SECONDS
        time.sleep(Job._POLL_INTERVAL_SECONDS)
    return self

  @property
  def state(self):
    """ Describe the state of a Job.

    Returns: A string describing the job's state.
    """
    state = 'in progress'
    if self.is_complete:
      if self.failed:
        state = 'failed with error: %s' % str(self._fatal_error)
      elif self._errors:
        state = 'completed with some non-fatal errors'
      else:
        state = 'completed'
    return state

  def __repr__(self):
    """ Get the notebook representation for the job. """
    return 'Job %s %s' % (self._job_id, self.state)

  @staticmethod
  def _wait(jobs, timeout, return_when):
    # If a single job is passed in, make it an array for consistency
    if isinstance(jobs, Job):
      jobs = [jobs]
    elif len(jobs) == 0:
      return jobs

    wait_on_one = return_when == concurrent.futures.FIRST_COMPLETED
    completed = []
    while True:
      if timeout is not None:
        timeout -= Job._POLL_INTERVAL_SECONDS

      done = [job for job in jobs if job.is_complete]

      if len(done):
        completed.extend(done)
        for job in done:
          jobs.remove(job)
        if wait_on_one or len(jobs) == 0:
          return completed

      if timeout is not None and timeout < 0:
        return completed

      # Need to block for some time. Favor using concurrent.futures.wait if possible
      # as it can return early if a (thread) job is ready; else fall back to time.sleep.
      futures = [job._future for job in jobs if job._future]
      if len(futures) == 0:
        time.sleep(Job._POLL_INTERVAL_SECONDS)
      else:
        concurrent.futures.wait(futures, timeout=Job._POLL_INTERVAL_SECONDS,
                                return_when=return_when)

  @staticmethod
  def wait_any(jobs, timeout=None):
    """ Return when at least one of the specified jobs has completed or timeout expires.

    Args:
      jobs: a Job or list of Jobs to wait on.
      timeout: a timeout in seconds to wait for. None (the default) means no timeout.
    Returns:
      A list of the jobs that have now completed or None if there were no jobs.

    """
    return Job._wait(jobs, timeout, concurrent.futures.FIRST_COMPLETED)

  @staticmethod
  def wait_all(jobs, timeout=None):
    """ Return when at all of the specified jobs have completed or timeout expires.

    Args:
      jobs: a Job or list of Jobs to wait on.
      timeout: a timeout in seconds to wait for. None (the default) means no timeout.
    Returns:
      A list of the jobs that have now completed or None if there were no jobs.
    """
    return Job._wait(jobs, timeout, concurrent.futures.ALL_COMPLETED)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Google Cloud Platform library - Internal Helpers."""

from ._async import async, async_function, async_method
from ._gcp_job import GCPJob
from ._http import Http, RequestException
from ._iterator import Iterator
from ._job import Job, JobError
from ._json_encoder import JSONEncoder
from ._lru_cache import LRUCache
from ._lambda_job import LambdaJob
from ._utils import print_exception_with_last_stack, get_item, compare_datetimes, \
                    pick_unused_port, is_http_running_on, gcs_copy_file



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations under
# the License.

"""Groups for the Google Monitoring API."""

from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import object

import collections
import fnmatch

import pandas

import datalab.context

from . import _utils


class Groups(object):
  """Represents a list of Stackdriver groups for a project."""

  _DISPLAY_HEADERS = ('Group ID', 'Group name', 'Parent ID', 'Parent name',
                      'Is cluster', 'Filter')

  def __init__(self, project_id=None, context=None):
    """Initializes the Groups for a Stackdriver project.

    Args:
      project_id: An optional project ID or number to override the one provided
          by the context.
      context: An optional Context object to use instead of the global default.
    """
    self._context = context or datalab.context.Context.default()
    self._project_id = project_id or self._context.project_id
    self._client = _utils.make_client(project_id, context)
    self._group_dict = None

  def list(self, pattern='*'):
    """Returns a list of groups that match the filters.

    Args:
      pattern: An optional pattern to filter the groups based on their display
          name. This can include Unix shell-style wildcards. E.g.
          ``"Production*"``.

    Returns:
      A list of Group objects that match the filters.
    """
    if self._group_dict is None:
      self._group_dict = collections.OrderedDict(
          (group.id, group) for group in self._client.list_groups())

    return [group for group in self._group_dict.values()
            if fnmatch.fnmatch(group.display_name, pattern)]

  def as_dataframe(self, pattern='*', max_rows=None):
    """Creates a pandas dataframe from the groups that match the filters.

    Args:
      pattern: An optional pattern to further filter the groups. This can
          include Unix shell-style wildcards. E.g. ``"Production *"``,
          ``"*-backend"``.
      max_rows: The maximum number of groups to return. If None, return all.

    Returns:
      A pandas dataframe containing matching groups.
    """
    data = []
    for i, group in enumerate(self.list(pattern)):
      if max_rows is not None and i >= max_rows:
        break
      parent = self._group_dict.get(group.parent_id)
      parent_display_name = '' if parent is None else parent.display_name
      data.append([
          group.id, group.display_name, group.parent_id,
          parent_display_name, group.is_cluster, group.filter])

    return pandas.DataFrame(data, columns=self._DISPLAY_HEADERS)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements BigQuery HTTP API wrapper."""
from __future__ import absolute_import
from __future__ import unicode_literals
from past.builtins import basestring
from builtins import object

import datalab.utils
import datalab.bigquery


class Api(object):
  """A helper class to issue BigQuery HTTP requests."""

  # TODO(nikhilko): Use named placeholders in these string templates.
  _ENDPOINT = 'https://www.googleapis.com/bigquery/v2'
  _JOBS_PATH = '/projects/%s/jobs/%s'
  _QUERIES_PATH = '/projects/%s/queries/%s'
  _DATASETS_PATH = '/projects/%s/datasets/%s'
  _TABLES_PATH = '/projects/%s/datasets/%s/tables/%s%s'
  _TABLEDATA_PATH = '/projects/%s/datasets/%s/tables/%s%s/data'

  _DEFAULT_PAGE_SIZE = 1024
  _DEFAULT_TIMEOUT = 60000

  def __init__(self, context):
    """Initializes the BigQuery helper with context information.

    Args:
      context: a Context object providing project_id and credentials.
    """
    self._credentials = context.credentials
    self._project_id = context.project_id

  @property
  def project_id(self):
    """The project_id associated with this API client."""
    return self._project_id

  def jobs_insert_load(self, source, table_name, append=False, overwrite=False, create=False,
                       source_format='CSV', field_delimiter=',', allow_jagged_rows=False,
                       allow_quoted_newlines=False, encoding='UTF-8', ignore_unknown_values=False,
                       max_bad_records=0, quote='"', skip_leading_rows=0):
    """ Issues a request to load data from GCS to a BQ table

    Args:
      source: the URL of the source bucket(s). Can include wildcards, and can be a single
          string argument or a list.
      table_name: a tuple representing the full name of the destination table.
      append: if True append onto existing table contents.
      overwrite: if True overwrite existing table contents.
      create: if True, create the table if it doesn't exist
      source_format: the format of the data; default 'CSV'. Other options are DATASTORE_BACKUP
          or NEWLINE_DELIMITED_JSON.
      field_delimiter: The separator for fields in a CSV file. BigQuery converts the string to
          ISO-8859-1 encoding, and then uses the first byte of the encoded string to split the data
          as raw binary (default ',').
      allow_jagged_rows: If True, accept rows in CSV files that are missing trailing optional
          columns; the missing values are treated as nulls (default False).
      allow_quoted_newlines: If True, allow quoted data sections in CSV files that contain newline
          characters (default False).
      encoding: The character encoding of the data, either 'UTF-8' (the default) or 'ISO-8859-1'.
      ignore_unknown_values: If True, accept rows that contain values that do not match the schema;
          the unknown values are ignored (default False).
      max_bad_records: The maximum number of bad records that are allowed (and ignored) before
          returning an 'invalid' error in the Job result (default 0).
      quote: The value used to quote data sections in a CSV file; default '"'. If your data does
          not contain quoted sections, set the property value to an empty string. If your data
          contains quoted newline characters, you must also enable allow_quoted_newlines.
      skip_leading_rows: A number of rows at the top of a CSV file to skip (default 0).
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._JOBS_PATH % (table_name.project_id, ''))
    if isinstance(source, basestring):
      source = [source]
    write_disposition = 'WRITE_EMPTY'
    if overwrite:
      write_disposition = 'WRITE_TRUNCATE'
    if append:
      write_disposition = 'WRITE_APPEND'
    data = {
        'kind': 'bigquery#job',
        'configuration': {
            'load': {
                'sourceUris': source,
                'destinationTable': {
                    'projectId': table_name.project_id,
                    'datasetId': table_name.dataset_id,
                    'tableId': table_name.table_id
                },
                'createDisposition': 'CREATE_IF_NEEDED' if create else 'CREATE_NEVER',
                'writeDisposition': write_disposition,
                'sourceFormat': source_format,
                'ignoreUnknownValues': ignore_unknown_values,
                'maxBadRecords': max_bad_records,
            }
        }
    }
    if source_format == 'CSV':
      load_config = data['configuration']['load']
      load_config.update({
          'fieldDelimiter': field_delimiter,
          'allowJaggedRows': allow_jagged_rows,
          'allowQuotedNewlines': allow_quoted_newlines,
          'quote': quote,
          'encoding': encoding,
          'skipLeadingRows': skip_leading_rows
      })

    return datalab.utils.Http.request(url, data=data, credentials=self._credentials)

  def jobs_insert_query(self, sql, code=None, imports=None, table_name=None, append=False,
                        overwrite=False, dry_run=False, use_cache=True, batch=True,
                        allow_large_results=False, table_definitions=None, dialect=None,
                        billing_tier=None):
    """Issues a request to insert a query job.

    Args:
      sql: the SQL string representing the query to execute.
      code: code for Javascript UDFs, if any.
      imports: a list of GCS URLs containing additional Javascript UDF support code, if any.
      table_name: None for an anonymous table, or a name parts tuple for a long-lived table.
      append: if True, append to the table if it is non-empty; else the request will fail if table
          is non-empty unless overwrite is True.
      overwrite: if the table already exists, truncate it instead of appending or raising an
          Exception.
      dry_run: whether to actually execute the query or just dry run it.
      use_cache: whether to use past query results or ignore cache. Has no effect if destination is
          specified.
      batch: whether to run this as a batch job (lower priority) or as an interactive job (high
          priority, more expensive).
      allow_large_results: whether to allow large results (slower with some restrictions but
          can handle big jobs).
      table_definitions: a list of JSON external table definitions for any external tables
          referenced in the query.
      dialect : {'legacy', 'standard'}, default 'legacy'
          'legacy' : Use BigQuery's legacy SQL dialect.
          'standard' : Use BigQuery's standard SQL (beta), which is
          compliant with the SQL 2011 standard.
      billing_tier: Limits the billing tier for this job. Queries that have resource
          usage beyond this tier will fail (without incurring a charge). If unspecified, this
          will be set to your project default. This can also be used to override your
          project-wide default billing tier on a per-query basis.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._JOBS_PATH % (self._project_id, ''))

    if dialect is None:
        dialect = datalab.bigquery.Dialect.default().bq_dialect

    data = {
        'kind': 'bigquery#job',
        'configuration': {
            'query': {
                'query': sql,
                'useQueryCache': use_cache,
                'allowLargeResults': allow_large_results,
                'useLegacySql': dialect == 'legacy'
            },
            'dryRun': dry_run,
            'priority': 'BATCH' if batch else 'INTERACTIVE',
        },
    }

    query_config = data['configuration']['query']

    resources = []
    if code:
      resources.extend([{'inlineCode': fragment} for fragment in code])

    if imports:
      resources.extend([{'resourceUri': uri} for uri in imports])

    query_config['userDefinedFunctionResources'] = resources

    if table_definitions:
      query_config['tableDefinitions'] = table_definitions

    if table_name:
      query_config['destinationTable'] = {
          'projectId': table_name.project_id,
          'datasetId': table_name.dataset_id,
          'tableId': table_name.table_id
      }
      if append:
        query_config['writeDisposition'] = "WRITE_APPEND"
      elif overwrite:
        query_config['writeDisposition'] = "WRITE_TRUNCATE"

    if billing_tier:
        query_config['maximumBillingTier'] = billing_tier

    return datalab.utils.Http.request(url, data=data, credentials=self._credentials)

  def jobs_query_results(self, job_id, project_id, page_size, timeout, start_index=0):
    """Issues a request to the jobs/getQueryResults method.

    Args:
      job_id: the id of job from a previously executed query.
      project_id: the project id to use to fetch the results; use None for the default project.
      page_size: limit to the number of rows to fetch.
      timeout: duration (in milliseconds) to wait for the query to complete.
      start_index: the index of the row (0-based) at which to start retrieving the page of result
          rows.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    if timeout is None:
      timeout = Api._DEFAULT_TIMEOUT
    if project_id is None:
      project_id = self._project_id

    args = {
        'maxResults': page_size,
        'timeoutMs': timeout,
        'startIndex': start_index
    }
    url = Api._ENDPOINT + (Api._QUERIES_PATH % (project_id, job_id))
    return datalab.utils.Http.request(url, args=args, credentials=self._credentials)

  def jobs_get(self, job_id, project_id=None):
    """Issues a request to retrieve information about a job.

    Args:
      job_id: the id of the job
      project_id: the project id to use to fetch the results; use None for the default project.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    if project_id is None:
      project_id = self._project_id
    url = Api._ENDPOINT + (Api._JOBS_PATH % (project_id, job_id))
    return datalab.utils.Http.request(url, credentials=self._credentials)

  def datasets_insert(self, dataset_name, friendly_name=None, description=None):
    """Issues a request to create a dataset.

    Args:
      dataset_name: the name of the dataset to create.
      friendly_name: (optional) the friendly name for the dataset
      description: (optional) a description for the dataset
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._DATASETS_PATH % (dataset_name.project_id, ''))
    data = {
        'kind': 'bigquery#dataset',
        'datasetReference': {
            'projectId': dataset_name.project_id,
            'datasetId': dataset_name.dataset_id
        },
    }
    if friendly_name:
      data['friendlyName'] = friendly_name
    if description:
      data['description'] = description
    return datalab.utils.Http.request(url, data=data, credentials=self._credentials)

  def datasets_delete(self, dataset_name, delete_contents):
    """Issues a request to delete a dataset.

    Args:
      dataset_name: the name of the dataset to delete.
      delete_contents: if True, any tables in the dataset will be deleted. If False and the
          dataset is non-empty an exception will be raised.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._DATASETS_PATH % dataset_name)
    args = {}
    if delete_contents:
      args['deleteContents'] = True
    return datalab.utils.Http.request(url, method='DELETE', args=args,
                                      credentials=self._credentials, raw_response=True)

  def datasets_update(self, dataset_name, dataset_info):
    """Updates the Dataset info.

    Args:
      dataset_name: the name of the dataset to update as a tuple of components.
      dataset_info: the Dataset resource with updated fields.
    """
    url = Api._ENDPOINT + (Api._DATASETS_PATH % dataset_name)
    return datalab.utils.Http.request(url, method='PUT', data=dataset_info,
                                  credentials=self._credentials)

  def datasets_get(self, dataset_name):
    """Issues a request to retrieve information about a dataset.

    Args:
      dataset_name: the name of the dataset
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._DATASETS_PATH % dataset_name)
    return datalab.utils.Http.request(url, credentials=self._credentials)

  def datasets_list(self, project_id=None, max_results=0, page_token=None):
    """Issues a request to list the datasets in the project.

    Args:
      project_id: the project id to use to fetch the results; use None for the default project.
      max_results: an optional maximum number of tables to retrieve.
      page_token: an optional token to continue the retrieval.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    if project_id is None:
      project_id = self._project_id
    url = Api._ENDPOINT + (Api._DATASETS_PATH % (project_id, ''))

    args = {}
    if max_results != 0:
      args['maxResults'] = max_results
    if page_token is not None:
      args['pageToken'] = page_token

    return datalab.utils.Http.request(url, args=args, credentials=self._credentials)

  def tables_get(self, table_name):
    """Issues a request to retrieve information about a table.

    Args:
      table_name: a tuple representing the full name of the table.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._TABLES_PATH % table_name)
    return datalab.utils.Http.request(url, credentials=self._credentials)

  def tables_list(self, dataset_name, max_results=0, page_token=None):
    """Issues a request to retrieve a list of tables.

    Args:
      dataset_name: the name of the dataset to enumerate.
      max_results: an optional maximum number of tables to retrieve.
      page_token: an optional token to continue the retrieval.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT +\
        (Api._TABLES_PATH % (dataset_name.project_id, dataset_name.dataset_id, '', ''))

    args = {}
    if max_results != 0:
      args['maxResults'] = max_results
    if page_token is not None:
      args['pageToken'] = page_token

    return datalab.utils.Http.request(url, args=args, credentials=self._credentials)

  def tables_insert(self, table_name, schema=None, query=None, friendly_name=None,
                    description=None):
    """Issues a request to create a table or view in the specified dataset with the specified id.
       A schema must be provided to create a Table, or a query must be provided to create a View.

    Args:
      table_name: the name of the table as a tuple of components.
      schema: the schema, if this is a Table creation.
      query: the query, if this is a View creation.
      friendly_name: an optional friendly name.
      description: an optional description.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + \
        (Api._TABLES_PATH % (table_name.project_id, table_name.dataset_id, '', ''))

    data = {
        'kind': 'bigquery#table',
        'tableReference': {
            'projectId': table_name.project_id,
            'datasetId': table_name.dataset_id,
            'tableId': table_name.table_id
        }
    }
    if schema:
      data['schema'] = {'fields': schema}
    if query:
      data['view'] = {'query': query}
    if friendly_name:
      data['friendlyName'] = friendly_name
    if description:
      data['description'] = description

    return datalab.utils.Http.request(url, data=data, credentials=self._credentials)

  def tabledata_insert_all(self, table_name, rows):
    """Issues a request to insert data into a table.

    Args:
      table_name: the name of the table as a tuple of components.
      rows: the data to populate the table, as a list of dictionaries.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._TABLES_PATH % table_name) + "/insertAll"

    data = {
      'kind': 'bigquery#tableDataInsertAllRequest',
      'rows': rows
    }

    return datalab.utils.Http.request(url, data=data, credentials=self._credentials)

  def tabledata_list(self, table_name, start_index=None, max_results=None, page_token=None):
    """ Retrieves the contents of a table.

    Args:
      table_name: the name of the table as a tuple of components.
      start_index: the index of the row at which to start retrieval.
      max_results: an optional maximum number of rows to retrieve.
      page_token: an optional token to continue the retrieval.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._TABLEDATA_PATH % table_name)
    args = {}
    if start_index:
      args['startIndex'] = start_index
    if max_results:
      args['maxResults'] = max_results
    if page_token is not None:
      args['pageToken'] = page_token
    return datalab.utils.Http.request(url, args=args, credentials=self._credentials)

  def table_delete(self, table_name):
    """Issues a request to delete a table.

    Args:
      table_name: the name of the table as a tuple of components.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._TABLES_PATH % table_name)
    return datalab.utils.Http.request(url, method='DELETE', credentials=self._credentials,
                                      raw_response=True)

  def table_extract(self, table_name, destination, format='CSV', compress=True,
                    field_delimiter=',', print_header=True):
    """Exports the table to GCS.

    Args:
      table_name: the name of the table as a tuple of components.
      destination: the destination URI(s). Can be a single URI or a list.
      format: the format to use for the exported data; one of CSV, NEWLINE_DELIMITED_JSON or AVRO.
          Defaults to CSV.
      compress: whether to compress the data on export. Compression is not supported for
          AVRO format. Defaults to False.
      field_delimiter: for CSV exports, the field delimiter to use. Defaults to ','
      print_header: for CSV exports, whether to include an initial header line. Default true.
    Returns:
      A parsed result object.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._JOBS_PATH % (table_name.project_id, ''))
    if isinstance(destination, basestring):
      destination = [destination]
    data = {
      # 'projectId': table_name.project_id, # Code sample shows this but it is not in job
      # reference spec. Filed as b/19235843
        'kind': 'bigquery#job',
        'configuration': {
            'extract': {
                'sourceTable': {
                    'projectId': table_name.project_id,
                    'datasetId': table_name.dataset_id,
                    'tableId': table_name.table_id,
                },
                'compression': 'GZIP' if compress else 'NONE',
                'fieldDelimiter': field_delimiter,
                'printHeader': print_header,
                'destinationUris': destination,
                'destinationFormat': format,
            }
        }
    }
    return datalab.utils.Http.request(url, data=data, credentials=self._credentials)

  def table_update(self, table_name, table_info):
    """Updates the Table info.

    Args:
      table_name: the name of the table to update as a tuple of components.
      table_info: the Table resource with updated fields.
    """
    url = Api._ENDPOINT + (Api._TABLES_PATH % table_name)
    return datalab.utils.Http.request(url, method='PUT', data=table_info,
                                      credentials=self._credentials)



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements CSV file exploration"""
from __future__ import absolute_import
from __future__ import unicode_literals


try:
  import IPython
  import IPython.core.magic
  import IPython.core.display
except ImportError:
  raise Exception('This module can only be loaded in ipython.')

import pandas as pd

import datalab.data

from . import _commands
from . import _utils


@IPython.core.magic.register_line_cell_magic
def csv(line, cell=None):
  parser = _commands.CommandParser.create('csv')

  view_parser = parser.subcommand('view',
                                  'Browse CSV files without providing a schema. ' +
                                  'Each value is considered string type.')
  view_parser.add_argument('-i', '--input',
                           help='Path of the input CSV data', required=True)
  view_parser.add_argument('-n', '--count',
                           help='The number of lines to browse from head, default to 5.')
  view_parser.add_argument('-P', '--profile', action='store_true',
                           default=False, help='Generate an interactive profile of the data')
  view_parser.set_defaults(func=_view)

  return _utils.handle_magic_line(line, cell, parser)


def _view(args, cell):
  csv = datalab.data.Csv(args['input'])
  num_lines = int(args['count'] or 5)
  headers = None
  if cell:
    ipy = IPython.get_ipython()
    config = _utils.parse_config(cell, ipy.user_ns)
    if 'columns' in config:
      headers = [e.strip() for e in config['columns'].split(',')]
  df = pd.DataFrame(csv.browse(num_lines, headers))
  if args['profile']:
    # TODO(gram): We need to generate a schema and type-convert the columns before this
    # will be useful for CSV
    return _utils.profile_df(df)
  else:
    return IPython.core.display.HTML(df.to_html(index=False))



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Google Cloud Platform library - BigQuery Functionality."""
from __future__ import absolute_import

from ._csv_options import CSVOptions
from ._dataset import Dataset, Datasets
from ._dialect import Dialect
from ._federated_table import FederatedTable
from ._job import Job
from ._query import Query
from ._query_job import QueryJob
from ._query_results_table import QueryResultsTable
from ._query_stats import QueryStats
from ._sampling import Sampling
from ._schema import Schema
from ._table import Table, TableMetadata
from ._udf import UDF
from ._utils import TableName, DatasetName
from ._view import View


def wait_any(jobs, timeout=None):
  """ Return when at least one of the specified jobs has completed or timeout expires.

  Args:
    jobs: a list of Jobs to wait on.
    timeout: a timeout in seconds to wait for. None (the default) means no timeout.
  Returns:
    Once at least one job completes, a list of all completed jobs.
    If the call times out then an empty list will be returned.

  """
  return Job.wait_any(jobs, timeout)


def wait_all(jobs, timeout=None):
  """ Return when all of the specified jobs have completed or timeout expires.

  Args:
    jobs: a single Job or list of Jobs to wait on.
    timeout: a timeout in seconds to wait for. None (the default) means no timeout.
  Returns:
    A list of completed Jobs. If the call timed out this will be shorter than the
    list of jobs supplied as a parameter.
  """
  return Job.wait_all(jobs, timeout)



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.


import google.cloud.ml.features as features
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
import pandas_profiling
from plotly.graph_objs import Histogram, Scatter, Scatter3d
from plotly.offline import iplot
from plotly import tools
import seaborn as sns
import tempfile

import datalab.utils

try:
  import IPython.core.display
except ImportError:
  raise Exception('This module can only be loaded in ipython.')


class DataSet(object):
  """Represents a dataset that can be explored through its 'analyze()' function.
     The data need to be able to fit in memory.
  """

  def __init__(self, feature_set, data_paths, format='csv'):
    """Initializes an instance of DataSet.

    Args:
      feature_set: A feature_set describing the data. The feature_set provides data types
          (for example, csv), column names, schema, data transformers, etc.
          This is the same class used in CloudML preprocessing.
      data_paths: A dictionary with {name: path} pair. All data need to have the same schema.
      format: the format of the data, currently only 'csv' or 'tsv'.

    Raises: Exception if data_paths is not a dictionary
            Exception if the format is not csv or tsv
    """
    self._feature_set = feature_set
    if not isinstance(data_paths, dict):
      raise Exception('Expect "data_paths" to be a dictionary.')
    self._data_paths = data_paths
    if format == 'csv':
      self._delimiter = ','
    elif format=='tsv':
      self._delimiter = '\t'
    else:
      raise Exception('Unsupported format "%s"' % format)
    self._dataframes = {}
    self._raw_dataframes = {}
    self._concatenated_data_frame = None
    self._concatenated_raw_data_frame = None
    self._target_name = None
    self._key_name = None

  def _get_dataframe_type(self, column):
    if isinstance(column, features.NumericFeatureColumn):
      return np.float64
    if isinstance(column, features.TargetFeatureColumn) and column.is_numeric:
      return np.float64
    return str

  def _is_categorical_column(self, column):
    if isinstance(column, features.CategoricalFeatureColumn):
      return True
    if isinstance(column, features.TargetFeatureColumn) and not column.is_numeric:
      return True
    return False

  def _transform_data(self, df):
    df = df.copy(deep=True)
    for name, value in type(self._feature_set).__dict__.iteritems():
      for column in (value if (type(value) == list or type(value) == tuple) else [value]):
        if self._is_categorical_column(column):
          concatenated_column = self._concatenated_raw_data_frame[column.name]
          all_categories = concatenated_column.astype('category').cat.categories
          df[column.name] = pd.Categorical(df[column.name], categories=all_categories)
        if isinstance(column, features.NumericFeatureColumn):
          concatenated_column = self._concatenated_raw_data_frame[column.name]
          # Simulate metadata so we can create a transformer from CloudML features registry.
          transform_info = {
            'type': 'numeric',
            'transform': column.transform,
          }
          transform_info[column.transform] = column.transform_args
          transform_info['max'] = max(concatenated_column)
          transform_info['min'] = min(concatenated_column)
          transformer = features._registries.transformation_registry \
              .get_transformer(transform_info)
          transformed = [transformer.transform(x)[0] for x in df[column.name]]
          if column.transform == 'discretize':
            # Transformed data contains a one_of_k list so need to convert it back to index.
            # Categories needs to be num_of_buckets+2 to match the transformer behavior,
            # where it creates a smaller-than-min and a greater-than-max buckets.
            df[column.name] = pd.Series(pd.Categorical(transformed,
                                                       categories=range(transformer._buckets+2)))
          else:
            # TODO(qimingj): It is supposed to work with most transformers but still need to
            #                test them if new transformers become available.
            df[column.name] = transformed
    return df

  def _load_to_dataframes(self):
    if self._concatenated_raw_data_frame is not None:
      return  # Already loaded.

    # Step 1: Get schema from feature_set class.
    schema = {}
    for name, value in type(self._feature_set).__dict__.iteritems():
      for column in (value if (type(value) == list or type(value) == tuple) else [value]):
        if issubclass(type(column), features.FeatureColumn):
          if isinstance(column, features.TargetFeatureColumn):
            self._target_name = column.name
          if isinstance(column, features.KeyFeatureColumn):
            self._key_name = column.name
          data_type = self._get_dataframe_type(column)
          schema[column.name] = data_type
    if self._target_name is None:
      raise Exception('No target column found from feature_set')

    # Step 2: Load all non-text data into raw dataframes.
    for name, data_path in self._data_paths.iteritems():
      local_file = data_path
      if data_path.startswith('gs://'):
        local_file = tempfile.mktemp()
        datalab.utils.gcs_copy_file(data_path, local_file)
      self._raw_dataframes[name] = pd.read_csv(local_file,
                                               names=type(self._feature_set).csv_columns,
                                               dtype=schema,
                                               delimiter=self._delimiter,
                                               skipinitialspace=True)
      if data_path.startswith('gs://'):
        os.remove(local_file)
    self._concatenated_raw_data_frame = pd.concat(self._raw_dataframes.values())

    # Step 3: Transform the data.
    for name, raw_df in self._raw_dataframes.iteritems():
      self._dataframes[name] = self._transform_data(raw_df)
    self._concatenated_data_frame = pd.concat(self._dataframes.values())

  def _get_numeric_values(self, df, column_name):
    if str(df[column_name].dtype) == 'category':
      return df[column_name].cat.codes
    else:
      return df[column_name].values

  def _create_dummy_trace(self, x, y):
    # Dummy trace is needed for scatter plot to a) set the same x and y ranges across multiple
    # subplots, b) the categorical labels are sorted in the same way across multiple subplots
    # (the order of the categories depend on the order they appear in the data).
    # For a given axis, if it is categorical data, we draw one point for each category.
    # If it is numeric data, we draw min and max. Usually on x and y axises we don't have same
    # number of points, so we will pad one axis data.
    # Note: This needs to go away if plotly python supports setting ranges and specifying
    # category order across subplots.
    if str(self._concatenated_data_frame[x].dtype) == 'category':
      categories = self._concatenated_data_frame[x].cat.categories
      x_dummy = list(categories)
    else:
      x_dummy = [min(self._concatenated_data_frame[x]), max(self._concatenated_data_frame[x])]
    if str(self._concatenated_data_frame[y].dtype) == 'category':
      categories = self._concatenated_data_frame[y].cat.categories
      y_dummy = list(categories)
    else:
      y_dummy = [min(self._concatenated_data_frame[y]), max(self._concatenated_data_frame[y])]
    if len(x_dummy) > len(y_dummy):
      y_dummy = y_dummy + [y_dummy[-1]]*(len(x_dummy)-len(y_dummy))
    if len(x_dummy) < len(y_dummy):
      x_dummy = x_dummy + [x_dummy[-1]]*(len(y_dummy)-len(x_dummy))

    scatter_dummy = Scatter(
      x=x_dummy,
      y=y_dummy,
      showlegend=False,
      opacity=0, # Make it invisible.
      hoverinfo='none',
    )
    return scatter_dummy

  def _histogram(self, names, x):
    concatenated_numeric_values = self._get_numeric_values(self._concatenated_data_frame, x)
    start = min(concatenated_numeric_values)
    end = max(concatenated_numeric_values)
    size = 1 if str(self._concatenated_data_frame[x].dtype) == 'category' \
        else (max(concatenated_numeric_values) - min(concatenated_numeric_values)) / 10.0
    fig = tools.make_subplots(rows=1, cols=len(names), print_grid=False)
    histogram_index = 1
    for name in names:
      df = self._dataframes[name]
      numeric_values = self._get_numeric_values(df, x)
      text = df[x].cat.categories if str(df[x].dtype) == 'category' else None
      histogram = Histogram(
        name=name,
        x=numeric_values,
        xbins=dict(
          start=start,
          end=end,
          size=size,
        ),
        text=text,
      )
      fig.append_trace(histogram, 1, histogram_index)
      fig.layout['xaxis' + str(histogram_index)].title = x
      fig.layout['xaxis' + str(histogram_index)].range = [start, end]
      fig.layout['yaxis' + str(histogram_index)].title = 'count'
      histogram_index += 1
    fig.layout.width = min(500 * len(names), 1200)
    fig.layout.height = 500
    iplot(fig)

  def _scatter_plot(self, names, x, y, color):
    showscale = True if str(self._concatenated_data_frame[color].dtype) != 'category' else False
    cmin = min(self._get_numeric_values(self._concatenated_data_frame, color))
    cmax = max(self._get_numeric_values(self._concatenated_data_frame, color))
    fig = tools.make_subplots(rows=1, cols=len(names), print_grid=False)
    scatter_index = 1
    scatter_dummy = self._create_dummy_trace(x, y)
    for name in names:
      df = self._dataframes[name]
      text = ["x=%s y=%s target=%s" % (str(a),str(b),str(t)) for a,b,t
              in zip(df[x], df[y], df[color])]
      scatter = Scatter(
        name=name,
        x=df[x],
        y=df[y],
        mode='markers',
        text=text,
        hoverinfo='text',
        marker=dict(
          color=self._get_numeric_values(df, color),
          colorscale='Viridis',
          showscale=showscale,
          cmin=cmin,
          cmax=cmax,
        )
      )
      # Add dummy trace to set same ranges and categorical orders on subplots.
      fig.append_trace(scatter_dummy, 1, scatter_index)
      fig.append_trace(scatter, 1, scatter_index)
      fig.layout['xaxis' + str(scatter_index)].title = x
      fig.layout['yaxis' + str(scatter_index)].title = y
      scatter_index += 1
    fig.layout.width = min(500 * len(names), 1200)
    fig.layout.height = 500
    iplot(fig)

  def _scatter3d_plot(self, names, x, y, z, color):
    showscale = True if str(self._concatenated_data_frame[color].dtype) != 'category' else False
    cmin = min(self._get_numeric_values(self._concatenated_data_frame, color))
    cmax = max(self._get_numeric_values(self._concatenated_data_frame, color))
    specs = [[{'is_3d':True}]*len(self._dataframes)]
    fig = tools.make_subplots(rows=1, cols=len(names), specs=specs, print_grid=False)
    scatter3d_index = 1
    for name in names:
      df = self._dataframes[name]
      text = ["x=%s y=%s z=%s, target=%s" % (str(a),str(b),str(c),str(t)) for a,b,c,t
              in zip(df[x], df[y], df[z], df[color])]
      scatter3d = Scatter3d(
        name=name,
        x=df[x],
        y=df[y],
        z=df[z],
        mode='markers',
        text=text,
        hoverinfo='text',
        marker=dict(
          color=self._get_numeric_values(df, color),
          colorscale='Viridis',
          showscale=showscale,
          cmin=cmin,
          cmax=cmax,
        )
      )
      fig.append_trace(scatter3d, 1, scatter3d_index)
      fig.layout['scene' + str(scatter3d_index)].xaxis.title = x
      fig.layout['scene' + str(scatter3d_index)].yaxis.title = y
      fig.layout['scene' + str(scatter3d_index)].zaxis.title = z
      scatter3d_index += 1
    fig.layout.width = min(500 * len(names), 1200)
    fig.layout.height = 500
    iplot(fig)

  def _plot_x(self, names, x):
    self._histogram(names, x)
    if x != self._target_name:
      self._scatter_plot(names, x, self._target_name, self._target_name)

  def _plot_xy(self, names, x, y):
    self._scatter_plot(names, x, y, self._target_name)

  def _plot_xyz(self, names, x, y, z):
    self._scatter3d_plot(names, x, y, z, self._target_name)

  def profile(self, names=None, columns=None):
    """Print profiles of the dataset.

    Args:
      names: the names of the data to plot. Such as ['train']. If None, all data in the datasets
          will be used.
      columns: The list of column names to plot correlations. If None, all numeric columns
          will be used.
    """
    self._load_to_dataframes()
    if names is None:
      names = self._raw_dataframes.keys()
    html = ''
    for name in names:
      df = self._raw_dataframes[name]
      html += '<br/><br/><p style="text-align:center"><font size="6">' + \
              '<b>%s</b></font></p><br/><br/>' % name
      if columns is not None:
        df = df[columns]
      html += pandas_profiling.ProfileReport(df).html.replace('bootstrap', 'nonexistent')
    IPython.core.display.display_html(IPython.core.display.HTML(html))

  def analyze(self, names=None, columns=None):
    """Analyze the data and report results in IPython output cell. The results are based
           on preprocessed data as described in feature_set.

    Args:
      names: the names of the data to plot. Such as ['train']. If None, all data in the datasets
          will be used.
      columns: The list of names of columns to analyze. If None, all numeric columns
          will be used.
          If one column is provided, displays a scatter plot between the column and target
          column, and a histogram of the column.
          If two columns are provided, displays a scatter plot between them,
          colored by target column.
          If three columns are provided, displays a 3d scatter plot between them,
          colored by target column.

    Raises:
      Exception if any column names are not found in the data or the columns are text.
      Exception if columns are greater than 3 or less than 1.
    """
    self._load_to_dataframes()
    if columns is None:
      columns = [x for x in self._concatenated_data_frame 
          if str(self._concatenated_data_frame[x].dtype) != 'object']
    if len(columns) > 3 or len(columns) < 1:
      raise 'Found %d columns. ' % len(columns) + \
          'Use "columns" parameter to specify one, two or three columns.'
    for column_name in columns:
      if column_name not in self._concatenated_data_frame:
        raise Exception('Cannot find column "%s"' % column_name)
      if str(self._concatenated_data_frame[column_name].dtype) == 'object':
        raise Exception('Cannot analyze text column "%s"' % column_name)

    if names is None:
      names = self._dataframes.keys()
    if len(columns) == 1:
      self._plot_x(names, columns[0])
    elif len(columns) == 2:
      self._plot_xy(names, columns[0], columns[1])
    elif len(columns) == 3:
      self._plot_xyz(names, columns[0], columns[1], columns[2])

  def to_dataframes(self):
    """Get the transformed data as a DataFrames

    Returns: the transformed data in {name: dataframe} dictionary.
    """
    self._load_to_dataframes()
    return self._dataframes

  def plot(self, names=None, columns=None):
    """Plot correlation graphs on the specified columns, in n*n grids.

    Args:
      names: the names of the data to plot. Such as ['train']. If None, all data in the datasets
          will be used.
      columns: The list of column names to plot correlations. If None, all numeric columns
          will be used.
    """
    self.to_dataframes()
    is_classification = (str(self._concatenated_data_frame[self._target_name].dtype) == 'category')
    if columns is not None and self._target_name not in columns and is_classification:
      columns.append(self._target_name)
    if names is None:
      names = self._dataframes.keys()

    for name in names:
      df_correlation = self._dataframes[name].copy(deep=True)
      if self._key_name is not None:
        del df_correlation[self._key_name]
      if columns is not None:
        df_correlation = df_correlation[columns]
      for k in df_correlation.columns:
        if str(df_correlation[k].dtype) == 'object' or str(df_correlation[k].dtype) == 'category':
          if k != self._target_name:
            # pairplot only works with numeric columns
            del df_correlation[k]
        else:
          # pairplot does not deal with missing values well. For now fillna(0).
          df_correlation[k] = df_correlation[k].fillna(0)
      sns.set(style="ticks", color_codes=True)
      if is_classification:
        # pairplot doesn't like categories with all numbers
        df_correlation[self._target_name] = map(lambda x: 'target ' + str(x), df_correlation[self._target_name])
        sns.pairplot(df_correlation, hue=self._target_name, dropna=True)
      else:
        sns.pairplot(df_correlation, dropna=True)
      plt.suptitle(name)
      plt.show()



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.


from __future__ import absolute_import

from . import _bigquery




# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

import yaml

import google.cloud.ml as ml


class Metadata(object):
  """Helper class for parsing and serving feature metadata.
  """

  def __init__(self, metadata_path):
    """Initializes an instance of Metadata.

    Args:
      metadata_path: metadata file path. Can be local or GCS path.
    """
    self._metadata_path = metadata_path
    self._labels = {}

  def get_classification_label(self, label_index):
    """Get classification label given a label index.

    Args:
      label_index: the index of label.

    Returns:
      The classification label, or label_index if the metadata is not for classification.

    Raises:
      Exception if metadata is malformed.
    """
    if len(self._labels) == 0:
      with ml.util._file.open_local_or_gcs(self._metadata_path, 'r') as f:
        metadata = yaml.load(f)
        if 'columns' not in metadata:
          raise Exception('Invalid metadata. No columns found.')
        for column_name, column in metadata['columns'].iteritems():
          scenario = column.get('scenario', None)
          # classification is the old name and now is called discrete.
          if scenario == 'classification' or scenario == 'discrete':
            if 'items' not in column:
              raise Exception('Invalid metadata. No "items" found for "%s"' % column_name)
            for label, index in column['items'].iteritems():
              self._labels[index] = label
            break
          elif scenario is not None:
            return label_index # target column found but not classification
      if len(self._labels) == 0:
        raise Exception('Invalid metadata. No classification labels found.')
    return self._labels[label_index]

  def get_target_name_and_scenario(self):
    """Get name of the target feature and scenario.

    Returns:
      Name of the target feature or scenario

    Raises:
      Exception if metadata is malformed.
    """
    with ml.util._file.open_local_or_gcs(self._metadata_path, 'r') as f:
      metadata = yaml.load(f)
    if 'features' not in metadata or 'columns' not in metadata:
      raise Exception('Invalid metadata. No features or columns found.')
    target_column_name, scenario = None, None
    for column_name, column in metadata['columns'].iteritems():
      if 'scenario' in column:
        target_column_name, scenario = column_name, column['scenario']
        break
    if target_column_name is None:
      raise Exception('Invalid metadata. No target found in columns.')
    for feature_name, feature in metadata['features'].iteritems():
      if feature['columns'][0] == target_column_name:
        return feature_name, scenario
    raise Exception('Invalid metadata. No target found in features.')



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements Storage HTTP API wrapper."""
from __future__ import absolute_import
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import object

import urllib.request, urllib.parse, urllib.error
import datalab.context
import datalab.utils


class Api(object):
  """A helper class to issue Storage HTTP requests."""

  # TODO(nikhilko): Use named placeholders in these string templates.
  _ENDPOINT = 'https://www.googleapis.com/storage/v1'
  _DOWNLOAD_ENDPOINT = 'https://www.googleapis.com/download/storage/v1'
  _UPLOAD_ENDPOINT = 'https://www.googleapis.com/upload/storage/v1'
  _BUCKET_PATH = '/b/%s'
  _OBJECT_PATH = '/b/%s/o/%s'
  _OBJECT_COPY_PATH = '/b/%s/o/%s/copyTo/b/%s/o/%s'

  _MAX_RESULTS = 100

  def __init__(self, context):
    """Initializes the Storage helper with context information.

    Args:
      context: a Context object providing project_id and credentials.
    """
    self._credentials = context.credentials
    self._project_id = context.project_id

  @property
  def project_id(self):
    """The project_id associated with this API client."""
    return self._project_id

  def buckets_insert(self, bucket, project_id=None):
    """Issues a request to create a new bucket.

    Args:
      bucket: the name of the bucket.
      project_id: the project to use when inserting the bucket.
    Returns:
      A parsed bucket information dictionary.
    Raises:
      Exception if there is an error performing the operation.
    """
    args = {'project': project_id if project_id else self._project_id}
    data = {'name': bucket}

    url = Api._ENDPOINT + (Api._BUCKET_PATH % '')
    return datalab.utils.Http.request(url, args=args, data=data, credentials=self._credentials)

  def buckets_delete(self, bucket):
    """Issues a request to delete a bucket.

    Args:
      bucket: the name of the bucket.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._BUCKET_PATH % bucket)
    datalab.utils.Http.request(url, method='DELETE', credentials=self._credentials,
                               raw_response=True)

  def buckets_get(self, bucket, projection='noAcl'):
    """Issues a request to retrieve information about a bucket.

    Args:
      bucket: the name of the bucket.
      projection: the projection of the bucket information to retrieve.
    Returns:
      A parsed bucket information dictionary.
    Raises:
      Exception if there is an error performing the operation.
    """
    args = {'projection': projection}
    url = Api._ENDPOINT + (Api._BUCKET_PATH % bucket)
    return datalab.utils.Http.request(url, credentials=self._credentials, args=args)

  def buckets_list(self, projection='noAcl', max_results=0, page_token=None, project_id=None):
    """Issues a request to retrieve the list of buckets.

    Args:
      projection: the projection of the bucket information to retrieve.
      max_results: an optional maximum number of objects to retrieve.
      page_token: an optional token to continue the retrieval.
      project_id: the project whose buckets should be listed.
    Returns:
      A parsed list of bucket information dictionaries.
    Raises:
      Exception if there is an error performing the operation.
    """
    if max_results == 0:
      max_results = Api._MAX_RESULTS

    args = {'project': project_id if project_id else self._project_id, 'maxResults': max_results}
    if projection is not None:
      args['projection'] = projection
    if page_token is not None:
      args['pageToken'] = page_token

    url = Api._ENDPOINT + (Api._BUCKET_PATH % '')
    return datalab.utils.Http.request(url, args=args, credentials=self._credentials)

  def object_download(self, bucket, key, start_offset=0, byte_count=None):
    """Reads the contents of an object as text.

    Args:
      bucket: the name of the bucket containing the object.
      key: the key of the object to be read.
      start_offset: the start offset of bytes to read.
      byte_count: the number of bytes to read. If None, it reads to the end.
    Returns:
      The text content within the object.
    Raises:
      Exception if the object could not be read from.
    """
    args = {'alt': 'media'}
    headers = {}
    if start_offset > 0 or byte_count is not None:
      header = 'bytes=%d-' % start_offset
      if byte_count is not None:
        header += '%d' % byte_count
      headers['Range'] = header
    url = Api._DOWNLOAD_ENDPOINT + (Api._OBJECT_PATH % (bucket, Api._escape_key(key)))
    return datalab.utils.Http.request(url, args=args, headers=headers,
                                      credentials=self._credentials, raw_response=True)

  def object_upload(self, bucket, key, content, content_type):
    """Writes text content to the object.

    Args:
      bucket: the name of the bucket containing the object.
      key: the key of the object to be written.
      content: the text content to be written.
      content_type: the type of text content.
    Raises:
      Exception if the object could not be written to.
    """
    args = {'uploadType': 'media', 'name': key}
    headers = {'Content-Type': content_type}

    url = Api._UPLOAD_ENDPOINT + (Api._OBJECT_PATH % (bucket, ''))
    return datalab.utils.Http.request(url, args=args, data=content, headers=headers,
                                      credentials=self._credentials, raw_response=True)

  def objects_copy(self, source_bucket, source_key, target_bucket, target_key):
    """Updates the metadata associated with an object.

    Args:
      source_bucket: the name of the bucket containing the source object.
      source_key: the key of the source object being copied.
      target_bucket: the name of the bucket that will contain the copied object.
      target_key: the key of the copied object.
    Returns:
      A parsed object information dictionary.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._OBJECT_COPY_PATH % (source_bucket, Api._escape_key(source_key),
                                                    target_bucket, Api._escape_key(target_key)))
    return datalab.utils.Http.request(url, method='POST', credentials=self._credentials)

  def objects_delete(self, bucket, key):
    """Deletes the specified object.

    Args:
      bucket: the name of the bucket.
      key: the key of the object within the bucket.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._OBJECT_PATH % (bucket, Api._escape_key(key)))
    datalab.utils.Http.request(url, method='DELETE', credentials=self._credentials,
                               raw_response=True)

  def objects_get(self, bucket, key, projection='noAcl'):
    """Issues a request to retrieve information about an object.

    Args:
      bucket: the name of the bucket.
      key: the key of the object within the bucket.
      projection: the projection of the object to retrieve.
    Returns:
      A parsed object information dictionary.
    Raises:
      Exception if there is an error performing the operation.
    """
    args = {}
    if projection is not None:
      args['projection'] = projection

    url = Api._ENDPOINT + (Api._OBJECT_PATH % (bucket, Api._escape_key(key)))
    return datalab.utils.Http.request(url, args=args, credentials=self._credentials)

  def objects_list(self, bucket, prefix=None, delimiter=None, projection='noAcl', versions=False,
                   max_results=0, page_token=None):
    """Issues a request to retrieve information about an object.

    Args:
      bucket: the name of the bucket.
      prefix: an optional key prefix.
      delimiter: an optional key delimiter.
      projection: the projection of the objects to retrieve.
      versions: whether to list each version of a file as a distinct object.
      max_results: an optional maximum number of objects to retrieve.
      page_token: an optional token to continue the retrieval.
    Returns:
      A parsed list of object information dictionaries.
    Raises:
      Exception if there is an error performing the operation.
    """
    if max_results == 0:
      max_results = Api._MAX_RESULTS

    args = {'maxResults': max_results}
    if prefix is not None:
      args['prefix'] = prefix
    if delimiter is not None:
      args['delimiter'] = delimiter
    if projection is not None:
      args['projection'] = projection
    if versions:
      args['versions'] = 'true'
    if page_token is not None:
      args['pageToken'] = page_token

    url = Api._ENDPOINT + (Api._OBJECT_PATH % (bucket, ''))
    return datalab.utils.Http.request(url, args=args, credentials=self._credentials)

  def objects_patch(self, bucket, key, info):
    """Updates the metadata associated with an object.

    Args:
      bucket: the name of the bucket containing the object.
      key: the key of the object being updated.
      info: the metadata to update.
    Returns:
      A parsed object information dictionary.
    Raises:
      Exception if there is an error performing the operation.
    """
    url = Api._ENDPOINT + (Api._OBJECT_PATH % (bucket, Api._escape_key(key)))
    return datalab.utils.Http.request(url, method='PATCH', data=info, credentials=self._credentials)

  @staticmethod
  def _escape_key(key):
    # Disable the behavior to leave '/' alone by explicitly specifying the safe parameter.
    return urllib.parse.quote(key, safe='')

  @staticmethod
  def verify_permitted_to_read(gs_path):
    """Check if the user has permissions to read from the given path.

    Args:
      gs_path: the GCS path to check if user is permitted to read.
    Raises:
      Exception if user has no permissions to read.
    """
    # TODO(qimingj): Storage APIs need to be modified to allow absence of project
    #                or credential on Items. When that happens we can move the function
    #                to Items class.
    from . import _bucket
    bucket, prefix = _bucket.parse_name(gs_path)
    credentials = None
    if datalab.context.Context.is_signed_in():
      credentials = datalab.context._utils.get_credentials()
    args = {
        'maxResults': Api._MAX_RESULTS,
        'projection': 'noAcl'
    }
    if prefix is not None:
      args['prefix'] = prefix
    url = Api._ENDPOINT + (Api._OBJECT_PATH % (bucket, ''))
    try:
      datalab.utils.Http.request(url, args=args, credentials=credentials)
    except datalab.utils.RequestException as e:
      if e.status == 401:
        raise Exception('Not permitted to read from specified path. '
                        'Please sign in and make sure you have read access.')
      raise e



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

# To publish to PyPi use: python setup.py bdist_wheel upload -r pypi

import datetime
import sys
from setuptools import setup

if sys.version_info[0] == 2:
  import platform
  import pip

minor = datetime.datetime.now().strftime("%y%m%d%H%M")
version = '0.1.' + minor

setup(
  name='datalab',
  version=version,
  namespace_packages=['datalab'],
  packages=[
    'datalab.bigquery',
    'datalab.bigquery.commands',
    'datalab.context',
    'datalab.context.commands',
    'datalab.data',
    'datalab.data.commands',
    'datalab.kernel',
    'datalab.mlalpha',
    'datalab.mlalpha.commands',
    'datalab.notebook',
    'datalab.stackdriver',
    'datalab.stackdriver.commands',
    'datalab.stackdriver.monitoring',
    'datalab.storage',
    'datalab.storage.commands',
    'datalab.utils',
    'datalab.utils.commands'
  ],
  description='Google Cloud Datalab',
  author='Google',
  author_email='google-cloud-datalab-feedback@googlegroups.com',
  url='https://github.com/googledatalab/datalab',
  download_url='https://github.com/googledatalab/datalab/tarball/0.1',
  keywords=[
    'Google',
    'GCP',
    'GCS',
    'bigquery'
  ],
  license="Apache Software License",
  classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
  ],
  long_description="""\
Support package for Google Cloud Datalab. This provides cell magics and Python APIs
for accessing Google's Cloud Platform services such as Google BigQuery.
  """,
  install_requires=[
    'future==0.15.2',
    'futures==3.0.5',
    'google-cloud==0.19.0',
    'google-api-python-client==1.5.1',
    'seaborn==0.7.0',
    'plotly==1.12.5',
    'httplib2==0.9.2',
    'oauth2client==2.2.0',
    'psutil==4.3.0',
    'pandas>=0.17.1',
    'pandas-profiling>=1.0.0a2',
    'python-dateutil==2.5.0',
    'pytz>=2015.4',
    'pyyaml==3.11',
    'requests==2.9.1',
    'ipykernel==4.4.1',
  ],
  package_data={
    'datalab.notebook': [
        'static/bigquery.css',
        'static/bigquery.js',
        'static/charting.css',
        'static/charting.js',
        'static/job.css',
        'static/job.js',
        'static/element.js',
        'static/style.js',
        'static/visualization.js',
        'static/codemirror/mode/sql.js',
        'static/parcoords.js',
        'static/extern/d3.parcoords.js',
        'static/extern/d3.parcoords.css',
        'static/extern/sylvester.js',
      ]
  }
)

# for python2 only, install tensorflow and cloudml
if sys.version_info[0] == 2:
  tensorflow_path = None
  if platform.system() == 'Darwin':
    tensorflow_path = 'https://storage.googleapis.com/tensorflow/mac/cpu/tensorflow-0.11.0-py2-none-any.whl'
  elif platform.system() == 'Linux':
    if platform.linux_distribution()[0] == 'Ubuntu':
      tensorflow_path = 'https://storage.googleapis.com/tensorflow/linux/cpu/tensorflow-0.11.0-cp27-none-linux_x86_64.whl'
    elif platform.linux_distribution()[0] == 'Debian':
      tensorflow_path = 'https://storage.googleapis.com/tensorflow/linux/cpu/debian/jessie/tensorflow-0.11.0-cp27-none-linux_x86_64.whl'

  # install tensorflow
  if not tensorflow_path:
    print("""Warning: could not find tensorflow build for your OS.
    Please go to https://www.tensorflow.org/get_started/os_setup to see install options""")
  else:
    pip.main(['install', tensorflow_path])

  # install cloud ml sdk
  pip.main(['install', 'https://storage.googleapis.com/cloud-ml/sdk/cloudml.latest.tar.gz'])



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements Projects functionality."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import object

import datalab.utils

from . import _api
from . import _utils


# We could do this with the gcloud SDK. However, installing that while locked on oauth2.5
# introduces some ugliness; in particular we are stuck with v0.9 and need to work around:
#
#   https://github.com/GoogleCloudPlatform/gcloud-python/issues/1412
#
# and would have to put up with:
#
#    https://github.com/GoogleCloudPlatform/gcloud-python/issues/1570
#
# So we use the REST API instead.

class Project(object):
  """ Simple wrapper class for Cloud projects. """

  def __init__(self, api, id, number, name):
    self._api = api
    self._id = id
    self._number = number
    self._name = name

  @property
  def id(self):
    return self._id

  @property
  def name(self):
    return self._name

  @property
  def number(self):
    return self._number

  def __str__(self):
    return self._id


class Projects(object):
  """ Iterator class for enumerating the active projects accessible to the account. """

  def __init__(self, credentials=None):
    """ Initialize the Projects object.

    Args:
      credentials: the credentials for the account.
    """
    if credentials is None:
      credentials = _utils.get_credentials()
    self._api = _api.Api(credentials)

  def _retrieve_projects(self, page_token, count):
    try:
      list_info = self._api.projects_list(max_results=count,
                                          page_token=page_token)
    except Exception as e:
      raise e

    projects = list_info.get('projects', [])
    if len(projects):
      try:
        projects = [Project(self._api,
                             info['projectId'],
                             info['projectNumber'],
                             info['name'])
                    for info in projects if info['lifecycleState'] == 'ACTIVE']
      except KeyError:
        raise Exception('Unexpected response from server.')

    page_token = list_info.get('nextPageToken', None)
    return projects, page_token

  def __iter__(self):
    """ Returns an iterator for iterating through the Datasets in the project.
    """
    return iter(datalab.utils.Iterator(self._retrieve_projects))

  @staticmethod
  def get_default_id(credentials=None):
    """ Get default project id.

    Returns: the default project id if there is one, or None.
    """
    project_id = _utils.get_project_id()
    if project_id is None:
      projects, _ = Projects(credentials)._retrieve_projects(None, 2)
      if len(projects) == 1:
        project_id = projects[0].id
    return project_id

  @staticmethod
  def save_default_id(project_id):
    """ Save default project id to config so it will persist across kernels and Datalab runs.

    Args:
      project_id: the project_id to save.
    """
    _utils.save_project_id(project_id)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Google Cloud Platform library - BigQuery UDF Functionality."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import str
from builtins import object

import json


class UDF(object):
  """Represents a BigQuery UDF declaration.
  """

  @property
  def name(self):
    return self._name

  @property
  def imports(self):
    return self._imports

  @property
  def code(self):
    return self._code

  def __init__(self, inputs, outputs, name, implementation, support_code=None, imports=None):
    """Initializes a Function object from its pieces.

    Args:
      inputs: a list of string field names representing the schema of input.
      outputs: a list of name/type tuples representing the schema of the output.
      name: the name of the javascript function
      implementation: a javascript function implementing the logic.
      support_code: additional javascript code that the function can use.
      imports: a list of GCS URLs or files containing further support code.
    Raises:
      Exception if the name is invalid.
      """
    self._outputs = outputs
    self._name = name
    self._implementation = implementation
    self._support_code = support_code
    self._imports = imports
    self._code = UDF._build_js(inputs, outputs, name, implementation, support_code)

  @staticmethod
  def _build_js(inputs, outputs, name, implementation, support_code):
    """Creates a BigQuery SQL UDF javascript object.

    Args:
      inputs: a list of (name, type) tuples representing the schema of input.
      outputs: a list of (name, type) tuples representing the schema of the output.
      name: the name of the function
      implementation: a javascript function defining the UDF logic.
      support_code: additional javascript code that the function can use.
    """
    # Construct a comma-separated list of input field names
    # For example, field1,field2,...
    input_fields = json.dumps([f[0] for f in inputs])

    # Construct a json representation of the output schema
    # For example, [{'name':'field1','type':'string'},...]
    output_fields = [{'name': f[0], 'type': f[1]} for f in outputs]
    output_fields = json.dumps(output_fields, sort_keys=True)

    # Build the JS from the individual bits with proper escaping of the implementation
    if support_code is None:
      support_code = ''
    return ('{code}\n{name}={implementation};\n' +
            'bigquery.defineFunction(\'{name}\', {inputs}, {outputs}, {name});')\
                .format(code=support_code,
                        name=name,
                        implementation=implementation,
                        inputs=str(input_fields),
                        outputs=str(output_fields))




# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements HTTP API wrapper."""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import object

import datalab.utils


class Api(object):
  """A helper class to issue API HTTP requests to resource manager API."""

  _ENDPOINT = 'https://cloudresourcemanager.googleapis.com/v1'
  _PROJECT_PATH = '/projects/%s'
  _PROJECTS_PATH = '/projects'

  def __init__(self, credentials):
    self._credentials = credentials

  def projects_list(self, max_results=0, page_token=None):
    url = Api._ENDPOINT + Api._PROJECTS_PATH
    args = {}
    if max_results != 0:
      args['pageSize'] = max_results
    if page_token is not None:
      args['pageToken'] = page_token

    return datalab.utils.Http.request(url, args=args, credentials=self._credentials)

  def project_get(self, projectId):
    url = Api._ENDPOINT + (Api._PROJECT_PATH % projectId)
    return datalab.utils.Http.request(url, credentials=self._credentials)

